import os
import argparse
import xarray as xr
import numpy as np

# TCBench Libraries
from . import constants

# %%
parser = argparse.ArgumentParser()
parser.add_argument(
    "--input",
    help="The input .dat file",
    default="/work/FAC/FGSE/IDYST/tbeucler/default/milton/repos/alpha_bench/data/SHIPS_dats/5day/lsdiag_na_1982_2019_sat_ts.dat",
)
parser.add_argument(
    "--output_folder",
    help="The output folder path",
    default="/work/FAC/FGSE/IDYST/tbeucler/default/milton/repos/alpha_bench/data/SHIPS_netcdfs/",
)

args = parser.parse_args()

# Define the input and output paths
input_path = args.input
output_folder = args.output_folder
# %%
# Read the .dat file
with open(input_path, "r") as file:
    data_dict = {}
    ds = None
    for line in file:
        data = line.split()

        # Check to see if the header line has been reached
        if data[-1] == "HEAD":
            # read in the header data
            data_dict["shortname"] = data[0]
            data_dict["year"] = data[1][:2]
            data_dict["month"] = data[1][2:4]
            data_dict["day"] = data[1][4:6]
            data_dict["hour"] = data[2]
            data_dict["max_wind"] = data[3]
            data_dict["lat"] = data[4]
            data_dict["lon"] = data[5]
            data_dict["mslp"] = data[6]
            data_dict["ATCF_ID"] = data[7]
            data_dict["datavars"] = []

            # If the dataset is None, create a new dataset
            if ds is None:
                ds = xr.Dataset()
                ds.attrs["ATCF_ID"] = data_dict["ATCF_ID"]
                print(ds.attrs["ATCF_ID"], flush=True)
            # if the dataset exists, check if the ATCF_ID is the same
            elif ds.attrs["ATCF_ID"] != data_dict["ATCF_ID"]:

                # If the ATCF_ID is different, check if the file
                # already exists. If it does, print that it does.
                # If it does not, write the dataset to a file
                if os.path.exists(
                    os.path.join(output_folder, f"{ds.attrs['ATCF_ID']}.nc")
                ):
                    print(
                        f"File {ds.attrs['ATCF_ID']}.nc already exists. Skipping.",
                        flush=True,
                    )
                else:
                    # write the dataset to a file
                    ds.to_netcdf(
                        os.path.join(output_folder, f"{ds.attrs['ATCF_ID']}.nc"),
                        mode="w",
                    )
                ds = xr.Dataset()
                ds.attrs["ATCF_ID"] = data_dict["ATCF_ID"]
                print(ds.attrs["ATCF_ID"], flush=True)

        if data[-1] == "TIME":
            # These are the lead times
            data_dict["lead_time"] = np.array(data[0:-1]).astype(np.dtype("i1"))
            # When the lead times are used in the code, it will be necessary
            # to change the lead times to timedelta[h] format

        # check if the line corresponds to a known variable type
        if np.any(np.isin(data, list(constants.SHIPS_metadata.keys()))):
            varname_idx = np.where(
                np.isin(data, list(constants.SHIPS_metadata.keys()))
            )[0].item()
            varname = data[varname_idx]
            vardata = np.array(data[:varname_idx]).astype(float)
            vardata[vardata == 9999] = np.nan
            attrs = constants.SHIPS_metadata[varname]
            data_length = int(attrs.get("Data_Length", len(vardata)))
            vardata = vardata[:data_length]
            if "Info_Label" in list(attrs.keys()):
                attrs["Info_Data"] = data[-1]

            data_dict["datavars"].append({varname: {"data": vardata, "attrs": attrs}})

        # if "RSST" in data:
        #     # Read the data into a dictionary
        #     data_dict["rsst"] = {
        #         "data": np.array(data[0:-2]).astype(float),
        #         "info_data": data[-1],
        #     }

        # if "HIST" in data:
        #     print(data)

        if data[-1] == "LAST":
            # TODO: write the data to the xarray dataset
            temp_ds = xr.Dataset()
            if int(data_dict["year"]) >= 82:
                year = f'19{data_dict["year"]}'
            else:
                year = f'20{data_dict["year"]}'
            timestamp = np.datetime64(
                f'{year}-{data_dict["month"]}-{data_dict["day"]}T{data_dict["hour"]}'
            ).astype("datetime64[ns]")

            # Iterate over the data variables
            for var in data_dict["datavars"]:
                for key in var.keys():
                    if var[key]["attrs"].get("LeadTimeVar", True):
                        coords = {
                            "t": [timestamp],
                            "lead_time": data_dict["lead_time"][
                                -var[key]["data"].size :
                            ],  # Only take the lead times up to the length of the data
                        }
                    else:
                        coords = {
                            "t": [timestamp],
                            "levels": np.arange(1, len(var[key]["data"]) + 1),
                        }

                    if "LeadTimeVar" in list(var[key]["attrs"].keys()):
                        var[key]["attrs"].pop("LeadTimeVar")

                    temp_ds = temp_ds.assign(
                        {
                            key: xr.DataArray(
                                var[key]["data"][
                                    None, :
                                ],  # Add a new dimension to the data
                                # dims=coords.keys(),
                                coords=coords,
                                attrs=var[key]["attrs"],
                            )
                        }
                    )
                # Add the lat, lon, max wind and mslp as single level vars
                temp_ds = temp_ds.assign(
                    {
                        "lat": xr.DataArray(
                            [data_dict["lat"]],
                            coords={"t": [timestamp]},
                            attrs={"long_name": "Latitude", "units": "degrees_north"},
                        ),
                        "lon": xr.DataArray(
                            [data_dict["lon"]],
                            coords={"t": [timestamp]},
                            attrs={"long_name": "Longitude", "units": "degrees_west"},
                        ),
                        "max_wind": xr.DataArray(
                            [data_dict["max_wind"]],
                            coords={"t": [timestamp]},
                            attrs={"long_name": "Maximum Wind Speed", "units": "knots"},
                        ),
                        "mslp": xr.DataArray(
                            [data_dict["mslp"]],
                            coords={"t": [timestamp]},
                            attrs={
                                "long_name": "Mean Sea Level Pressure",
                                "units": "hPa",
                            },
                        ),
                    }
                )

            ds = xr.merge([ds, temp_ds])

            # Reset the data dictionary
            data_dict = {}

# %%



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 15 11:44:20 2022

Python script used to download ERA5 data from the CDS API into single files
with multiple pressure levels.

@author: mgomezd1
"""

import cdsapi
import os
import numpy as np
import argparse

parser = argparse.ArgumentParser(description="Download ERA5 data")

# read in the bash arguments
parser.add_argument(
    "--datavar", type=str, default="divergence", help="variable to download"
)
parser.add_argument("--year", type=int, default=2020, help="year to download")
parser.add_argument("--month", type=int, default=1, help="Month to download")

args = parser.parse_args()

folder_path = "/work/FAC/FGSE/IDYST/tbeucler/default/raw_data/ECMWF/ERA5/"

# load client interface
client = cdsapi.Client(timeout=86000, quiet=False, sleep_max=900, retry_max=1000)


datavars = [
    args.datavar,
]

pressure_level_vars = [
    "divergence",
    "fraction_of_cloud_cover",
    "geopotential",
    "ozone_mass_mixing_ratio",
    "potential_vorticity",
    "relative_humidity",
    "specific_cloud_ice_water_content",
    "specific_cloud_liquid_water_content",
    "specific_humidity",
    "specific_rain_water_content",
    "specific_snow_water_content",
    "temperature",
    "u_component_of_wind",
    "v_component_of_wind",
    "vertical_velocity",
    "vorticity",
]

single_level_vars = [
    "100m_u_component_of_wind",
    "100m_v_component_of_wind",
    "10m_u_component_of_neutral_wind",
    "10m_u_component_of_wind",
    "10m_v_component_of_neutral_wind",
    "10m_v_component_of_wind",
    "10m_wind_gust_since_previous_post_processing",
    "2m_dewpoint_temperature",
    "2m_temperature",
    "air_density_over_the_oceans",
    "angle_of_sub_gridscale_orography",
    "anisotropy_of_sub_gridscale_orography",
    "benjamin_feir_index",
    "boundary_layer_dissipation",
    "boundary_layer_height",
    "charnock",
    "clear_sky_direct_solar_radiation_at_surface",
    "cloud_base_height",
    "coefficient_of_drag_with_waves",
    "convective_available_potential_energy",
    "convective_inhibition",
    "convective_precipitation",
    "convective_rain_rate",
    "convective_snowfall",
    "convective_snowfall_rate_water_equivalent",
    "downward_uv_radiation_at_the_surface",
    "duct_base_height",
    "eastward_gravity_wave_surface_stress",
    "eastward_turbulent_surface_stress",
    "evaporation",
    "forecast_albedo",
    "forecast_logarithm_of_surface_roughness_for_heat",
    "forecast_surface_roughness",
    "free_convective_velocity_over_the_oceans",
    "friction_velocity",
    "geopotential",
    "gravity_wave_dissipation",
    "high_cloud_cover",
    "high_vegetation_cover",
    "ice_temperature_layer_1",
    "ice_temperature_layer_2",
    "ice_temperature_layer_3",
    "ice_temperature_layer_4",
    "instantaneous_10m_wind_gust",
    "instantaneous_eastward_turbulent_surface_stress",
    "instantaneous_large_scale_surface_precipitation_fraction",
    "instantaneous_moisture_flux",
    "instantaneous_northward_turbulent_surface_stress",
    "instantaneous_surface_sensible_heat_flux",
    "k_index",
    "lake_bottom_temperature",
    "lake_cover",
    "lake_depth",
    "lake_ice_depth",
    "lake_ice_temperature",
    "lake_mix_layer_depth",
    "lake_mix_layer_temperature",
    "lake_shape_factor",
    "lake_total_layer_temperature",
    "land_sea_mask",
    "large_scale_precipitation",
    "large_scale_precipitation_fraction",
    "large_scale_rain_rate",
    "large_scale_snowfall",
    "large_scale_snowfall_rate_water_equivalent",
    "leaf_area_index_high_vegetation",
    "leaf_area_index_low_vegetation",
    "low_cloud_cover",
    "low_vegetation_cover",
    "maximum_2m_temperature_since_previous_post_processing",
    "maximum_individual_wave_height",
    "maximum_total_precipitation_rate_since_previous_post_processing",
    "mean_boundary_layer_dissipation",
    "mean_convective_precipitation_rate",
    "mean_convective_snowfall_rate",
    "mean_direction_of_total_swell",
    "mean_direction_of_wind_waves",
    "mean_eastward_gravity_wave_surface_stress",
    "mean_eastward_turbulent_surface_stress",
    "mean_evaporation_rate",
    "mean_gravity_wave_dissipation",
    "mean_large_scale_precipitation_fraction",
    "mean_large_scale_precipitation_rate",
    "mean_large_scale_snowfall_rate",
    "mean_northward_gravity_wave_surface_stress",
    "mean_northward_turbulent_surface_stress",
    "mean_period_of_total_swell",
    "mean_period_of_wind_waves",
    "mean_potential_evaporation_rate",
    "mean_runoff_rate",
    "mean_sea_level_pressure",
    "mean_snow_evaporation_rate",
    "mean_snowfall_rate",
    "mean_snowmelt_rate",
    "mean_square_slope_of_waves",
    "mean_sub_surface_runoff_rate",
    "mean_surface_direct_short_wave_radiation_flux",
    "mean_surface_direct_short_wave_radiation_flux_clear_sky",
    "mean_surface_downward_long_wave_radiation_flux",
    "mean_surface_downward_long_wave_radiation_flux_clear_sky",
    "mean_surface_downward_short_wave_radiation_flux",
    "mean_surface_downward_short_wave_radiation_flux_clear_sky",
    "mean_surface_downward_uv_radiation_flux",
    "mean_surface_latent_heat_flux",
    "mean_surface_net_long_wave_radiation_flux",
    "mean_surface_net_long_wave_radiation_flux_clear_sky",
    "mean_surface_net_short_wave_radiation_flux",
    "mean_surface_net_short_wave_radiation_flux_clear_sky",
    "mean_surface_runoff_rate",
    "mean_surface_sensible_heat_flux",
    "mean_top_downward_short_wave_radiation_flux",
    "mean_top_net_long_wave_radiation_flux",
    "mean_top_net_long_wave_radiation_flux_clear_sky",
    "mean_top_net_short_wave_radiation_flux",
    "mean_top_net_short_wave_radiation_flux_clear_sky",
    "mean_total_precipitation_rate",
    "mean_vertical_gradient_of_refractivity_inside_trapping_layer",
    "mean_vertically_integrated_moisture_divergence",
    "mean_wave_direction",
    "mean_wave_direction_of_first_swell_partition",
    "mean_wave_direction_of_second_swell_partition",
    "mean_wave_direction_of_third_swell_partition",
    "mean_wave_period",
    "mean_wave_period_based_on_first_moment",
    "mean_wave_period_based_on_first_moment_for_swell",
    "mean_wave_period_based_on_first_moment_for_wind_waves",
    "mean_wave_period_based_on_second_moment_for_swell",
    "mean_wave_period_based_on_second_moment_for_wind_waves",
    "mean_wave_period_of_first_swell_partition",
    "mean_wave_period_of_second_swell_partition",
    "mean_wave_period_of_third_swell_partition",
    "mean_zero_crossing_wave_period",
    "medium_cloud_cover",
    "minimum_2m_temperature_since_previous_post_processing",
    "minimum_total_precipitation_rate_since_previous_post_processing",
    "minimum_vertical_gradient_of_refractivity_inside_trapping_layer",
    "model_bathymetry",
    "near_ir_albedo_for_diffuse_radiation",
    "near_ir_albedo_for_direct_radiation",
    "normalized_energy_flux_into_ocean",
    "normalized_energy_flux_into_waves",
    "normalized_stress_into_ocean",
    "northward_gravity_wave_surface_stress",
    "northward_turbulent_surface_stress",
    "ocean_surface_stress_equivalent_10m_neutral_wind_direction",
    "ocean_surface_stress_equivalent_10m_neutral_wind_speed",
    "peak_wave_period",
    "period_corresponding_to_maximum_individual_wave_height",
    "potential_evaporation",
    "precipitation_type",
    "runoff",
    "sea_ice_cover",
    "sea_surface_temperature",
    "significant_height_of_combined_wind_waves_and_swell",
    "significant_height_of_total_swell",
    "significant_height_of_wind_waves",
    "significant_wave_height_of_first_swell_partition",
    "significant_wave_height_of_second_swell_partition",
    "significant_wave_height_of_third_swell_partition",
    "skin_reservoir_content",
    "skin_temperature",
    "slope_of_sub_gridscale_orography",
    "snow_albedo",
    "snow_density",
    "snow_depth",
    "snow_evaporation",
    "snowfall",
    "snowmelt",
    "soil_temperature_level_1",
    "soil_temperature_level_2",
    "soil_temperature_level_3",
    "soil_temperature_level_4",
    "soil_type",
    "standard_deviation_of_filtered_subgrid_orography",
    "standard_deviation_of_orography",
    "sub_surface_runoff",
    "surface_latent_heat_flux",
    "surface_net_solar_radiation",
    "surface_net_solar_radiation_clear_sky",
    "surface_net_thermal_radiation",
    "surface_net_thermal_radiation_clear_sky",
    "surface_pressure",
    "surface_runoff",
    "surface_sensible_heat_flux",
    "surface_solar_radiation_downward_clear_sky",
    "surface_solar_radiation_downwards",
    "surface_thermal_radiation_downward_clear_sky",
    "surface_thermal_radiation_downwards",
    "temperature_of_snow_layer",
    "toa_incident_solar_radiation",
    "top_net_solar_radiation",
    "top_net_solar_radiation_clear_sky",
    "top_net_thermal_radiation",
    "top_net_thermal_radiation_clear_sky",
    "total_cloud_cover",
    "total_column_cloud_ice_water",
    "total_column_cloud_liquid_water",
    "total_column_ozone",
    "total_column_rain_water",
    "total_column_snow_water",
    "total_column_supercooled_liquid_water",
    "total_column_water",
    "total_column_water_vapour",
    "total_precipitation",
    "total_sky_direct_solar_radiation_at_surface",
    "total_totals_index",
    "trapping_layer_base_height",
    "trapping_layer_top_height",
    "type_of_high_vegetation",
    "type_of_low_vegetation",
    "u_component_stokes_drift",
    "uv_visible_albedo_for_diffuse_radiation",
    "uv_visible_albedo_for_direct_radiation",
    "v_component_stokes_drift",
    "vertical_integral_of_divergence_of_cloud_frozen_water_flux",
    "vertical_integral_of_divergence_of_cloud_liquid_water_flux",
    "vertical_integral_of_divergence_of_geopotential_flux",
    "vertical_integral_of_divergence_of_kinetic_energy_flux",
    "vertical_integral_of_divergence_of_mass_flux",
    "vertical_integral_of_divergence_of_moisture_flux",
    "vertical_integral_of_divergence_of_ozone_flux",
    "vertical_integral_of_divergence_of_thermal_energy_flux",
    "vertical_integral_of_divergence_of_total_energy_flux",
    "vertical_integral_of_eastward_cloud_frozen_water_flux",
    "vertical_integral_of_eastward_cloud_liquid_water_flux",
    "vertical_integral_of_eastward_geopotential_flux",
    "vertical_integral_of_eastward_heat_flux",
    "vertical_integral_of_eastward_kinetic_energy_flux",
    "vertical_integral_of_eastward_mass_flux",
    "vertical_integral_of_eastward_ozone_flux",
    "vertical_integral_of_eastward_total_energy_flux",
    "vertical_integral_of_eastward_water_vapour_flux",
    "vertical_integral_of_energy_conversion",
    "vertical_integral_of_kinetic_energy",
    "vertical_integral_of_mass_of_atmosphere",
    "vertical_integral_of_mass_tendency",
    "vertical_integral_of_northward_cloud_frozen_water_flux",
    "vertical_integral_of_northward_cloud_liquid_water_flux",
    "vertical_integral_of_northward_geopotential_flux",
    "vertical_integral_of_northward_heat_flux",
    "vertical_integral_of_northward_kinetic_energy_flux",
    "vertical_integral_of_northward_mass_flux",
    "vertical_integral_of_northward_ozone_flux",
    "vertical_integral_of_northward_total_energy_flux",
    "vertical_integral_of_northward_water_vapour_flux",
    "vertical_integral_of_potential_and_internal_energy",
    "vertical_integral_of_potential_internal_and_latent_energy",
    "vertical_integral_of_temperature",
    "vertical_integral_of_thermal_energy",
    "vertical_integral_of_total_energy",
    "vertically_integrated_moisture_divergence",
    "volumetric_soil_water_layer_1",
    "volumetric_soil_water_layer_2",
    "volumetric_soil_water_layer_3",
    "volumetric_soil_water_layer_4",
    "wave_spectral_directional_width",
    "wave_spectral_directional_width_for_swell",
    "wave_spectral_directional_width_for_wind_waves",
    "wave_spectral_kurtosis",
    "wave_spectral_peakedness",
    "wave_spectral_skewness",
    "zero_degree_level",
]

# Pressure levels to download according to our groups requirements
full_pressure = [
    "10",
    "20",
    "30",
    "50",
    "70",
    "100",
    "150",
    "200",
    "250",
    "300",
    "400",
    "500",
    "600",
    "700",
    "800",
    "850",
    "900",
    "925",
    "950",
    "975",
    "1000",
]

# Download every 3 hours, which is the maximum (interpolated) resolution for ibtracs
times = [
    "00:00",
    "03:00",
    "06:00",
    "09:00",
    "12:00",
    "15:00",
    "18:00",
    "21:00",
]

years = [
    args.year,
]

months = [
    args.month,
]

if args.datavar in pressure_level_vars:
    directory_path = os.path.join(folder_path, "PL", args.datavar)
elif args.datavar in single_level_vars:
    directory_path = os.path.join(folder_path, "SL", args.datavar)
else:
    raise ValueError(
        f"Variable {args.datavar} not recognized as a single or pressure level variable"
    )

if not os.path.exists(directory_path):
    os.makedirs(directory_path)

filename = f"ERA5_{args.year}-{args.month:02d}_{args.datavar}.nc"
target_path = os.path.join(directory_path, filename)
if os.path.exists(target_path):
    print(f"File {filename} already exists")
    exit()

# assert that the variable is in the list of available variables
assert (
    args.datavar in pressure_level_vars + single_level_vars
), f"Variable {args.datavar} not a recognized ERA5 variable"


data_params = {
    "product_type": "reanalysis",
    "format": "netcdf",
    "variable": datavars,
    "year": years,
    "month": months,
    "day": list(np.arange(1, 32, 1).astype(str)),
    "time": times,
}

if args.datavar in pressure_level_vars:
    data_origin = "reanalysis-era5-pressure-levels"
    data_params["pressure_level"] = full_pressure
else:
    data_origin = "reanalysis-era5-single-levels"

# print(data_origin, data_params, target_path)
client.retrieve(name=data_origin, request=data_params, target=target_path)



# %% Imports
# OS and IO
import os
import sys
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import dask.array as da

# Backend Libraries
import joblib as jl

from utils import toolbox, constants
from utils import data_lib as dlib

# %% Load the seasons to process

seasons = toolbox.get_TC_seasons(
    season_list=[*range(2016, 2021)],
    datadir_path="/work/FAC/FGSE/IDYST/tbeucler/default/raw_data/TCBench_alpha",
)

# %% Control flags

process = True
perfect_prog = True

# %% Process the tracks
input_samples = None
target_samples = None
for season, storms in seasons.items():
    print(f"Starting to process {season}. which contains {len(storms)} storms...")

    if process:
        # Load the data collection
        data_dir = (
            "/work/FAC/FGSE/IDYST/tbeucler/default/raw_data/AI-milton/panguweather"
        )
        dc = dlib.AI_Data_Collection(data_dir)

        # determine the number of processors that can be used
        n_jobs = jl.cpu_count()

        data_dir = (
            "/work/FAC/FGSE/IDYST/tbeucler/default/raw_data/AI-milton/panguweather"
        )
        dc = dlib.AI_Data_Collection(data_dir)

        # process the tracks
        jl.Parallel(n_jobs=n_jobs)(
            jl.delayed(storm.process_data_collection)(dc) for storm in storms
        )

    else:
        # for storm in storms:
        #     inputs = None
        #     outputs = None
        #     try:
        #         inputs, outputs, t, leads = storm.serve_ai_data()
        #     except Exception as e:
        #         print(f"Failed to process {str(storm)}")
        #         print(f"Error: {e}")
        #     if inputs is not None and outputs is not None:
        #         if input_samples is None:
        #             input_samples = inputs
        #             target_samples = outputs
        #         else:
        #             input_samples = da.vstack((input_samples, inputs))
        #             target_samples = da.vstack((target_samples, outputs))
        pass

# %%



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  9 10:53:01 2023

This library contains dictionaries and values useful as a reference for other
scripts used in TC Bench

@author: mgomezd1
"""

# %% Library Imports
import numpy as np
import pandas as pd


# %% Columns for track files
class track_cols:
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            if key != "__META":
                setattr(self, key, val)
            else:
                self.__metadata = val

    def get_dtypes(self):
        dtype_dict = {}

        for key, dtype in self.__dict__.items():
            if key[0] != "_":
                dtype_dict[key] = (
                    str
                    if (isinstance(dtype, dict) or np.issubdtype(dtype, np.datetime64))
                    else dtype
                )
        return dtype_dict

    def get_colnames(self):
        colnames = []
        for key, dtype in self.__dict__.items():
            if key[0] != "_":
                colnames.append(key)
        return colnames

    def get_datetime_cols(self):
        datetime_cols = []

        for key, dtype in self.__dict__.items():
            if dtype == np.datetime64:
                datetime_cols.append(key)
        return datetime_cols


# Track column data for tracks prepared by Dr. Saranya Ganesh Sudheesh
sgs_track_cols = track_cols(  # Basin Corresponts to ibtracs basin
    basin={
        "NA": "North Atlantic",
        "EP": "Eastern North Pacific",
        "WP": "Western North Pacific",
        "NI": "North Indian",
        "SI": "South Indian",
        "SP": "South Pacific",
        "SA": "South Atlantic",
    },
    # Region Corresponts to ibtracs subbasin
    region={
        "MM": "Missing",
        "CS": "Caribbean Sea",
        "GM": "Gulf of Mexico",
        "CP": "Central Pacific",
        "BB": "Bay of Bengal",
        "AS": "Arabian Sea",
        "WA": "Western Australia",
        "EA": "Eastern Australia",
    },
    # Name of the storm
    name=str,
    # Track step
    no=int,
    # Track point timestamp
    datetime=np.datetime64,
    # Intensity corresponds to ibtracs nature
    intensity={
        "DS": "Disturbance",
        "TS": "Tropical",
        "ET": "Extratropical",
        "SS": "Subtropical",
        "NR": "Not Reported",
        "MX": "Mixture",
    },
    # Coordinates
    lat=np.float16,
    lon=np.float16,
    # Metadata for constructing track objects
    __META={
        "UID": "SID",
        "COSMETIC_NAME": "NAME",
        "Y_coord": "lat",
        "X_coord": "lon",
        "TIME_coord": "datetime",
        "loader": pd.read_csv,
    },
)

# Track column data for ibtracs file
ibtracs_cols = track_cols(  # Storm ID number
    SID=str,
    # Year of the event
    SEASON=int,
    # Number of event in season CURRENTLY IGNORED
    # NUMBER = int
    # Basin
    BASIN={
        "NA": "North Atlantic",
        "EP": "Eastern North Pacific",
        "WP": "Western North Pacific",
        "NI": "North Indian",
        "SI": "South Indian",
        "SP": "South Pacific",
        "SA": "South Atlantic",
        "MM": "Missing",  # Note that MM should not appear in final ibtracs product
    },
    # Subbasin
    SUBBASIN={
        "MM": "Missing",
        "CS": "Caribbean Sea",
        "GM": "Gulf of Mexico",
        "CP": "Central Pacific",
        "BB": "Bay of Bengal",
        "AS": "Arabian Sea",
        "WA": "Western Australia",
        "EA": "Eastern Australia",
    },
    # Name of the storm provided by the agencies
    NAME=str,
    # Track point ISO timestamp
    ISO_TIME=np.datetime64,
    # IBTRACS Combined storm type
    NATURE={
        "DS": "Disturbance",
        "TS": "Tropical",
        "ET": "Extratropical",
        "SS": "Subtropical",
        "NR": "Not Reported",
        "MX": "Mixture",
    },
    # Coordinates
    LAT=np.float16,
    LON=np.float16,
    # US Automated Tropical Cyclone Forecasting System ID
    USA_ATCF_ID=str,
    # Intensity and Pressure
    USA_WIND=str,
    USA_PRES=str,
    # Metadata for constructing track objects
    __META={
        "UID": "SID",
        "COSMETIC_NAME": "NAME",
        "Y_coord": "LAT",
        "X_coord": "LON",
        "TIME_coord": "ISO_TIME",
        "loader": pd.read_csv,
        "ALT_ID": "USA_ATCF_ID",
        "WIND": "USA_WIND",
        "PRES": "USA_PRES",
        "SEASON": "SEASON",
    },
)

# Track column data for Meteo France La Reunion file
reunion_track_cols = track_cols(  # Storm Season
    SAISON=int,
    # Name of the storm
    NOM_CYC=str,
    # Seasonal number of the depression
    NUM_DEPR=int,
    # Storm ID
    ID=str,
    # Track point timestamp
    ANNEE=np.datetime64,
    MOIS=np.datetime64,
    JOUR=np.datetime64,
    HEURE_UTC=np.datetime64,
    # Coordinates
    LAT=np.float16,
    LON=np.float16,
    # Metadata for constructing track objects
    __META={
        "UID": "ID",
        "COSMETIC_NAME": "NOM_CYC",
        "Y_coord": "LAT",
        "X_coord": "LON",
        "TIME_coord": "date",
        "loader": "meteo_france",
    },
)

# %% Folder Structure Long Names
# The folder dictionary is used to expand the folder names to their full
# names. This is useful for the user to understand what the folder names
# mean.
# The ERA5 folder structure is organized by single level, pressure level, and
# calculated values.
data_store_names = {
    "SL": "Surface / Single Level",
    "PL": "Pressure Level",
    "CV": "Calculated Values",
}

# %% valid coordinate names
valid_coords = {
    "latitude": ["lat", "Lat"],
    "longitude": ["lon", "Lon"],
    "time": ["time", "Time", "t"],
    "level": ["Lev", "lev", "pressure", "Pressure", "isobaric"],
    "leadtime": ["lead", "Lead", "ldt", "LDT", "valid", "Valid"],
}

# default variable names
default_ships_vars = [
    "DELV",  # Intensity change
    "VMAX",  # Maximum sustained wind
    "MSLP",  # Minimum sea level pressure
    "lat",  # Latitude
    "CSST",  # Climatological sea surface temperature
    # "PSLV",  # Climatological sea level pressure
    "Z850",  # 850 hPa geopotential height
    "D200",  # 200 hPa divergence
    "EPOS",  # Surface - Environment theta_e difference, r = 200-800
    "RHMD",  # 700-500 hPa relative humidity
    "TWAC",  # Average symmetric tangential wind at 850 hPa (0-600km average)
    "G200",  # Averaged 200 hPa Temperature Perturbation (r=200-800km)
    "TADV",  # Averaged 850 to 700 hPa temperature advection, r=0-500km
    "SHGC",  # Generalized 850-200 hPa shear magnitude, vortex removed (r=0-500km)
    "LHRD",  # SHDC * sin(lat)
    "VSHR",  # VMAX*SHDC
    "T200",  # 200 hPa temperature
    "T250",  # 250 hPa temperature
    "SHDC",  # 850-200 hPa shear magnitude, vortex removed (r=0-500km)
]
# %% SHIPS Metadata
SHIPS_metadata = {
    "VMAX": {"long_name": "Maximum Surface Wind", "units": "kt"},
    "MSLP": {"long_name": "Minimum Sea Level Pressure", "units": "hPa"},
    "TYPE": {
        "long_name": "Storm type",
        "units": "None",
        "Notes": "0=wave, remnant low, dissipating low, 1=tropical, 2=subtropical, 3=extra-tropical. \n Note that the SHIPS variables are set to missing for all cases except type=1 or 2,\nsince these are not included in the SHIPS developmental sample for estimating the model coefficients.",
    },
    "HIST": {
        "long_name": "Storm history",
        "units": "6hr periods",
        "Notes": "The no. of 6 hour periods the storm max wind has been above 20, 25, ,…, 120 kt.",
        "LeadTimeVar": False,
    },
    "DELV": {
        "long_name": "Intensity change (see notes)",
        "units": "kt",
        "Notes": "-12 to 0, -6 to 0, 0 to 0, 0 to 6, ... 0 to 120 hr. If the storm crosses a major land mass during the time interval, value set to NaN (9999 in original dataset)",
        "LeadTimeVar": False,
    },
    "INCV": {
        "long_name": "Intensity change (see notes)",
        "units": "kt",
        "Notes": "-18 to -12, -12 to -6, ... 114 to 120 hr. Set to 9999 similar to DELV for land cases.",
        "LeadTimeVar": False,
    },
    "LAT": {"long_name": "Storm latitude", "units": "deg N *10", "Notes": "vs time"},
    "LON": {"long_name": "Storm longitude", "units": "deg W *10", "Notes": "vs time"},
    "CSST": {
        "long_name": "Climatological SST",
        "units": "deg C * 10",
        "Notes": "vs time",
    },
    "CD20": {
        "long_name": "Climatological depth of 20 deg C isotherm",
        "units": "m",
        "Notes": "from 2005-2010 NCODA analyses",
    },
    "CD26": {
        "long_name": "Climatological depth of 26 deg C isotherm",
        "units": "m",
        "Notes": "from 2005-2010 NCODA analyses",
    },
    "COHC": {
        "long_name": "Climatological ocean heat content",
        "units": "kJ/cm2",
        "Notes": "from 2005-2010 NCODA analyses",
    },
    "DTL": {
        "long_name": "Distance to nearest major land mass",
        "units": "km",
        "Notes": "vs time",
    },
    "OAGE": {
        "long_name": "Ocean Age",
        "units": "hr*10",
        "Notes": "The amount of time the area within 100 km of the storm center has been occupied by the storm along its track up to this point in time",
    },
    "NAGE": {
        "long_name": "Normalized Ocean Age",
        "units": "hr*10",
        "Notes": "Same as OAGE, but normalized by the maximum wind/100kt. If the max wind was a constant 100 kt over its past history, NAGE=OAGE.",
    },
    "RSST": {
        "long_name": "Reynolds Sea Surface Temperature",
        "units": "deg C*10",
        "Notes": "vs time",
        "Info_Label": "Age, in days, of the SST analysis used to estimate RSST",
    },
    "DSST": {
        "long_name": "Daily Reynolds Sea Surface Temperature",
        "units": "deg C*10",
        "Notes": "vs time",
        "Info_Label": "Age, in days, of the SST analysis used to estimate RSST",
    },
    "DSTA": {
        "long_name": "Spatially Averaged Daily Reynolds Sea Surface Temperature",
        "units": "deg C*10",
        "Notes": "vs time. spatially averaged over 5 points (storm center, + 50 km N, E, S and W of center)",
        "Info_Label": "Age, in days, of the SST analysis used to estimate RSST",
    },
    "PHCN": {
        "long_name": "Estimated ocean heat content",
        "units": "kJ/cm2",
        "Notes": "From climo OHC and current SST anomaly. Designed to fill in for RHCN when that is missing.",
    },
    "U200": {"long_name": "200 hPa zonal wind, r=200-800 km", "units": "kt *10"},
    "U20C": {"long_name": "200 hPa zonal wind, r=0-500 km", "units": "kt *10"},
    "V20C": {"long_name": "200 hPa meridional wind, r=0-500 km", "units": "kt *10"},
    "E000": {
        "long_name": "1000 hPa Potential Temperature (Theta_e)",
        "units": "Unit",
        "Notes": "",
    },
    "EPOS": {
        "long_name": "Surface - Environment Potential Temperature (theta_e) difference, r = 200-800",
        "units": "deg K*10",
        "Notes": "The average theta e difference between a parcel lifted from the surface and its environment \n(200-800 km average) vs time. Only positive differences are included in the average.",
    },
    "ENEG": {
        "long_name": "Negative Surface - Environment Potential Temperature (theta_e) difference, r = 200-800",
        "units": "Unit",
        "Notes": "The average theta e difference between a parcel lifted from the surface and its environment \n(200-800 km average) vs time. Only negative differences are included in the average.\nMinus sign not included.",
    },
    "EPSS": {
        "long_name": "Negative Surface - Saturated-Environment Potential Temperature (theta_e) difference, r = 200-800",
        "units": "Unit",
        "Notes": "The average theta_e difference between a parcel lifted from the surface and the saturated theta_e \nof its environment (200-800 km average) vs time. Only positive differences are included in the average.",
    },
    "ENSS": {
        "long_name": "Negative Surface - Saturated-Environment Potential Temperature (theta_e) difference, r = 200-800",
        "units": "Unit",
        "Notes": "The average theta_e difference between a parcel lifted from the surface and the saturated theta_e \nof its environment (200-800 km average) vs time. Only negative differences are included in the average.\nMinus sign not included.",
    },
    "RHLO": {
        "long_name": "850-700 hPa relative humidity, r=200-800 km",
        "units": "%",
        "Notes": "vs time",
    },
    "RHMD": {
        "long_name": "700-500 hPa relative humidity, r=200-800 km",
        "units": "%",
        "Notes": "vs time",
    },
    "RHHI": {
        "long_name": "500-300 hPa relative humidity, r=200-800 km",
        "units": "%",
        "Notes": "vs time",
    },
    "PSLV": {
        "long_name": "Pressure, Storm Center Motion, Deep Layer Wind, and Weights",
        "units": "See Notes",
        "Notes": f"""
                 First Column: Pressure of the center of mass of the layer where storm motion best matches environmental flow (hPa) (t=0 only)
                 Second Column: The observed zonal storm motion component (m/s *10)
                 Third Column: The observed meridional storm motion component (m/s *10)
                 Fourth Column: The observed zonal storm motion component (m/s *10) for the 1000 to 100 hPa mass weighted deep layer environmental wind
                 Fifth Column: The observed meridional storm motion component (m/s *10) for the 1000 to 100 hPa mass weighted deep layer environmental wind
                 Sixth Column: The observed zonal storm motion component (m/s *10) for the optimally weighted deep layer mean flow
                 Seventh Column: The observed meridional storm motion component (m/s *10) for the optimally weighted deep layer mean flow
                 Eighth Column: The parameter alpha that controls the constraint on the weights from being not too “far” from the deep layer mean weights (non-dimensional, *100)
                 Ninth Column: The optimal vertical weights for p=100 hPa. (non-dimensional *1000)
                 Tenth Column: The optimal vertical weights for p=150 hPa. (non-dimensional *1000)
                 Eleventh Column: The optimal vertical weights for p=200 hPa. (non-dimensional *1000)
                 Twelfth Column: The optimal vertical weights for p=250 hPa. (non-dimensional *1000)
                 Thirteenth Column: The optimal vertical weights for p=300 hPa. (non-dimensional *1000)
                 Fourteenth Column: The optimal vertical weights for p=400 hPa. (non-dimensional *1000)
                 Fifteenth Column: The optimal vertical weights for p=500 hPa. (non-dimensional *1000)
                 Sixteenth Column: The optimal vertical weights for p=700 hPa. (non-dimensional *1000)
                 Seventeenth Column: The optimal vertical weights for p=850 hPa. (non-dimensional *1000)
                 Eighteenth Column: The optimal vertical weights for p=1000 hPa. (non-dimensional *1000)""",
        "LeadTimeVar": False,
        "Data_Length": 18,
    },
    "Z850": {
        "long_name": "850 hPa vorticity (r=0-1000km)",
        "units": "sec-1 * 10**7",
        "Notes": "vs time",
    },
    "D200": {
        "long_name": "200 hPa divergence (r=0-1000km)",
        "units": "sec-1 * 10**7",
        "Notes": "vs time",
    },
    "REFC": {
        "long_name": "Relative eddy momentum flux convergence, 100-600 km avg",
        "units": "m/sec/day",
        "Notes": "vs time",
    },
    "PEFC": {
        "long_name": "Planetary eddy momentum flux convergence, 100-600 km avg",
        "units": "m/sec/day",
        "Notes": "vs time",
    },
    "T000": {
        "long_name": "1000 hPa temperature, 200-800 km average",
        "units": "dec C* 10",
        "Notes": "",
    },
    "R000": {
        "long_name": "1000 hPa relative humidity, 200-800 km average",
        "units": "%",
        "Notes": "",
    },
    "Z000": {
        "long_name": "1000 hPa height deviation",
        "units": "m",
        "Notes": "Height deviation from the U.S. standard atmosphere",
    },
    "TLAT": {
        "long_name": "850 hPa vortex center latitude",
        "units": "deg N*10",
        "Notes": "Vortex in the NCEP analysis",
    },
    "TLON": {
        "long_name": "850 hPa vortex center longitude",
        "units": "deg W*10",
        "Notes": "Vortex in the NCEP analysis",
    },
    "TWAC": {
        "long_name": "Average symmetric tangential wind at 850 hPa (0-600km average)",
        "units": "m/sec *10",
        "Notes": "from NCEP analysis",
    },
    "TWXC": {
        "long_name": "Maximum 850 hPa symmetric tangential wind",
        "units": "m/sec *10",
        "Notes": "from NCEP analysis",
    },
    "G150": {
        "long_name": "Averaged 150 hPa Temperature Perturbation (r=200-800km)",
        "units": "deg C*10",
        "Notes": """Perturbation due to the symmetric vortex calculated from the gradient thermal wind.
        Radius centered on input lat/lon (not always the model/analysis vortex position)""",
    },
    "G200": {
        "long_name": "Averaged 200 hPa Temperature Perturbation (r=200-800km)",
        "units": "deg C*10",
        "Notes": """Perturbation due to the symmetric vortex calculated from the gradient thermal wind.
        Radius centered on input lat/lon (not always the model/analysis vortex position)""",
    },
    "G250": {
        "long_name": "Averaged 250 hPa Temperature Perturbation (r=200-800km)",
        "units": "deg C*10",
        "Notes": """Perturbation due to the symmetric vortex calculated from the gradient thermal wind.
        Radius centered on input lat/lon (not always the model/analysis vortex position)""",
    },
    "V000": {
        "long_name": "Azimuthally averaged 1000 hPa tangential wind (r=500km)",
        "units": "m/sec *10",
        "Notes": "Azimuthally averaged at r=500 km from (TLAT,TLON). \nIf TLAT,TLON are not available, (LAT,LON) are used.",
    },
    "V850": {
        "long_name": "Azimuthally averaged 850 hPa tangential wind (r=500km)",
        "units": "m/sec *10",
        "Notes": "Azimuthally averaged at r=500 km from (TLAT,TLON). \nIf TLAT,TLON are not available, (LAT,LON) are used.",
    },
    "V500": {
        "long_name": "Azimuthally averaged 500 hPa tangential wind (r=500km)",
        "units": "m/sec *10",
        "Notes": "Azimuthally averaged at r=500 km from (TLAT,TLON). \nIf TLAT,TLON are not available, (LAT,LON) are used.",
    },
    "V300": {
        "long_name": "Azimuthally averaged 300 hPa tangential wind (r=500km)",
        "units": "m/sec *10",
        "Notes": "Azimuthally averaged at r=500 km from (TLAT,TLON). \nIf TLAT,TLON are not available, (LAT,LON) are used.",
    },
    "TGRD": {
        "long_name": "Averaged 850 to 700 hPa temperature gradient magnitude, r=0-500km",
        "units": "deg C per m*107",  # Is this right? Verify with @Mark and @Kate
        "Notes": "Estimated from the geostrophic thermal wind",
    },
    "TADV": {
        "long_name": "Averaged 850 to 700 hPa temperature advection, r=0-500km",
        "units": "deg per sec*106",  # Is this right? Verify with @Mark and @Kate
        "Notes": "Estimated from the geostrophic thermal wind",
    },
    "PENC": {
        "long_name": "Azimuthally averaged surface pressure at outer edge of vortex",
        "units": "(hPa-1000)*10",
        "Notes": "",
    },
    "SHRD": {
        "long_name": "850-200 hPa shear magnitude (200-800km)",
        "units": "kt *10",
        "Notes": "vs time",
    },
    "SHDC": {
        "long_name": "850-200 hPa shear magnitude, vortex removed (r=0-500km)",
        "units": "kt *10",
        "Notes": "vs time. Averaged from 0-500 km relative to 850 hPa vortex center",
    },
    "SDDC": {
        "long_name": "Heading of above shear vector",
        "units": "deg",
        "Notes": "Westerly shear has a value of 90 deg",
    },
    "SHRG": {
        "long_name": "Generalized 850-200 hPa shear magnitude",
        "units": "kt *10",
        "Notes": "vs time. Takes into account all levels from 1000 to 100 hPa",
    },
    "SHGC": {
        "long_name": "Generalized 850-200 hPa shear magnitude, vortex removed (r=0-500km)",
        "units": "kt *10",
        "Notes": "vs time. Takes into account all levels from 1000 to 100 hPa. \nAveraged from 0-500 km relative to 850 hPa vortex center",
    },
    "DIVC": {
        "long_name": "Centered 200 hPa divergence (r=0-1000km)",
        "units": "sec-1 * 10**7",
        "Notes": "vs time. Centered at 850 hPa vortex location",
    },
    "T150": {
        "long_name": "200 to 800 km area average 150 hPa temperature",
        "units": "deg C *10",
        "Notes": "versus time",
    },
    "T200": {
        "long_name": "200 to 800 km area average 200 hPa temperature",
        "units": "deg C *10",
        "Notes": "versus time",
    },
    "T250": {
        "long_name": "200 to 800 km area average 250 hPa temperature",
        "units": "deg C *10",
        "Notes": "versus time",
    },
    "SHTD": {  # How is this different from SHTS? SDDC? @Mark and @Kate
        "long_name": "Heading of above shear vector",
        "units": "deg",
        "Notes": "Westerly shear has a value of 90 deg.",
    },
    "SHRS": {
        "long_name": "850-500 hPa shear magnitude",
        "units": "kt *10",
        "Notes": "vs time",
    },
    "SHTS": {
        "long_name": "Heading of above shear vector",
        "units": "deg",  # verify with @Mark and @Kate
        "Notes": "",  # Should this say 90deg for Westerly shear? verify with @Mark and @Kate
    },
    "PENV": {
        "long_name": "200 to 800 km average surface pressure",
        "units": "(hPa-1000)*10",
        "Notes": "",
    },
    "VMPI": {
        "long_name": "Maximum potential intensity",
        "units": "kt",
        "Notes": "from Kerry Emanuel equation",
    },
    "VVAV": {
        "long_name": "Average surface parcel vertical velocity",
        "units": "m/s * 100",
        "Notes": f"""Average vertical velocity (0 to 15 km) of a parcel lifted from the surface
                 where entrainment, the ice phase and the condensate weight are accounted for.
                 Note: Moisture and temperature biases between the operational and reanalysis files
                 make this variable inconsistent in the 2001-2007 sample, compared 2000 and before.""",
    },
    "VMFX": {
        "long_name": "Density Weighted Average surface parcel vertical velocity",
        "units": "m/s * 100",
        "Notes": f"""Average vertical velocity (0 to 15 km) of a parcel lifted from the surface
                 where entrainment, the ice phase and the condensate weight are accounted for.
                 Density weighted vertical average.
                 Note: Moisture and temperature biases between the operational and reanalysis files
                 make this variable inconsistent in the 2001-2007 sample, compared 2000 and before.""",
    },
    "VVAC": {
        "long_name": "Average surface parcel vertical velocity, vortex removed, from soundings",
        "units": "m/s * 100",
        "Notes": f"""
                 Same as VVAV but with soundings from 0-500 km with GFS vortex removed
                 Average vertical velocity (0 to 15 km) of a parcel lifted from the surface
                 where entrainment, the ice phase and the condensate weight are accounted for.
                 Note: Moisture and temperature biases between the operational and reanalysis files
                 make this variable inconsistent in the 2001-2007 sample, compared 2000 and before.""",
    },
    "HE07": {
        "long_name": "1000-700 hPa Storm motion relative helicity, r=200-800 km",
        "units": "m^2/s^2",
        "Notes": "",
    },
    "HE05": {
        "long_name": "1000-500 hPa Storm motion relative helicity, r=200-800 km",
        "units": "m^2/s^2",
        "Notes": "",
    },
    "O500": {
        "long_name": "Average 500 hPa Pressure vertical velocity, r=0-1000 km ",
        "units": "(hPa/day",
        "Notes": "",
    },
    "O700": {
        "long_name": "Average 700 hPa Pressure vertical velocity, r=0-1000 km ",
        "units": "(hPa/day",
        "Notes": "",
    },
    "CFLX": {
        "long_name": "Dry air predictor",
        "units": "%",  # veriy with @Mark and @ Kate
        "Notes": "based on the difference in surface moisture flux between air with the \nobserved (GFS) RH value,and with RH of air mixed \nfrom 500 hPa to the surface.",
    },
    "MTPW": {
        "long_name": "Total Precipitable Water (TPW) predictors @ t=0",
        "units": "See Notes",
        "Notes": f"""
                     1)	0-200 km average TPW (mm * 10)
                     2)	0-200 km TPW standard deviation (mm * 10)
                     3)	200-400 km average TPW (mm * 10)
                     4)	200-400 km TPW standard deviation (mm * 10)
                     5)	400-600 km average TPW (mm * 10)
                     6)	400-600 km TPW standard deviation (mm * 10)
                     7)	600-800 km average TPW (mm * 10)
                     8)	600-800 km TPW standard deviation (mm * 10)
                     9)	800-1000 km average TPW (mm * 10)
                     10)	800-1000 km TPW standard deviation (mm * 10)
                     11)	0-400 km average TPW (mm * 10)
                     12)	0-400 km TPW standard deviation (mm * 10)
                     13)	0-600 km average TPW (mm * 10)
                     14)	0-600 km TPW standard deviation (mm * 10)
                     15)	0-800 km average TPW (mm * 10)
                     16)	0-800 km TPW standard deviation (mm * 10)
                     17)	0-1000 km average TPW (mm * 10)
                     18)	0-1000 km TPW standard deviation (mm * 10)
                     19)	%TPW less than 45 mm, r=0 to 500 km in 90 deg azimuthal quadrant centered on up-shear direction
                     20)	0-500 km averaged TPW (mm * 10) in 90 deg up-shear quadrant
                     21)	0-500 km average TPW (mm * 10)""",
        "LeadTimeVar": False,
        "Data_Length": 21,
    },
    "PW01": {
        "long_name": "0-200 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW02": {
        "long_name": "0-200 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW03": {
        "long_name": "200-400 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW04": {
        "long_name": "200-400 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW05": {
        "long_name": "400-600 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW06": {
        "long_name": "400-600 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW07": {
        "long_name": "600-800 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW08": {
        "long_name": "600-800 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW09": {
        "long_name": "800-1000 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW10": {
        "long_name": "800-1000 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW11": {"long_name": "0-400 km average TPW", "units": "mm*10", "Notes": ""},
    "PW12": {
        "long_name": "0-400 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW13": {
        "long_name": "0-600 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW14": {
        "long_name": "0-600 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW15": {"long_name": "0-800 km average TPW", "units": "mm*10", "Notes": ""},
    "PW16": {
        "long_name": "0-800 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW17": {"long_name": "0-1000 km average TPW", "units": "mm*10", "Notes": ""},
    "PW18": {
        "long_name": "0-1000 km TPW standard deviation",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "PW19": {
        "long_name": "%TPW less than 45 mm",
        "units": "Unit",
        "Notes": "r=0 to 500 km in 90 deg azimuthal quadrant centered on up-shear direction. Time dependent",
    },
    "PW20": {
        "long_name": "0-500 km averaged TPW ",
        "units": "mm*10",
        "Notes": "in 90 deg up-shear quadrant. Time dependent",
    },
    "PW21": {
        "long_name": "0-500 km average TPW",
        "units": "mm*10",
        "Notes": "Time dependent",
    },
    "IR00": {
        "long_name": "Predictors from GOES data",
        "units": "See Notes",
        "Notes": f""" Not time dependent. The 20 values in this record are as follows:
                    1) Time (hhmm) of the GOES image, relative to this case
                    2) Average GOES ch 4 brightness temp (deg C *10), r=0-200 km
                    3) Stan. Dev. of GOES BT (deg C*10), r=0-200 km
                    4) Same as 2) for r=100-300 km
                    5) Same as 3) for r=100-300 km
                    6) Percent area r=50-200 km of GOES ch 4 BT < -10 C
                    7) Same as 6 for BT < -20 C
                    8) Same as 6 for BT < -30 C
                    9) Same as 6 for BT < -40 C
                    10) Same as 6 for BT < -50 C
                    11) Same as 6 for BT < -60 C
                    12) max BT from 0 to 30 km radius (deg C*10)
                    13) avg BT from 0 to 30 km radius (deg C*10)
                    14) radius of max BT (km)
                    15) min BT from 20 to 120 km radius (deg C*10)
                    16) avg BT from 20 to 120 km radius (deg C*10)
                    17)  radius of min BT (km)
                    18-20) Variables need for storm size estimation""",
        "LeadTimeVar": False,
        "Data_Length": 20,
    },
    "IRXX": {
        "long_name": "Inferred Predictors substituting GOES data",
        "units": "See Notes",
        "Notes": f"""Not time dependent.
                     The same as IR00 but generated from other predictors (not satellite data).
                     These should only be used to fill in for missing IR00 if needed.
                    1) Time (hhmm) of the GOES image, relative to this case
                    2) Average GOES ch 4 brightness temp (deg C *10), r=0-200 km
                    3) Stan. Dev. of GOES BT (deg C*10), r=0-200 km
                    4) Same as 2) for r=100-300 km
                    5) Same as 3) for r=100-300 km
                    6) Percent area r=50-200 km of GOES ch 4 BT < -10 C
                    7) Same as 6 for BT < -20 C
                    8) Same as 6 for BT < -30 C
                    9) Same as 6 for BT < -40 C
                    10) Same as 6 for BT < -50 C
                    11) Same as 6 for BT < -60 C
                    12) max BT from 0 to 30 km radius (deg C*10)
                    13) avg BT from 0 to 30 km radius (deg C*10)
                    14) radius of max BT (km)
                    15) min BT from 20 to 120 km radius (deg C*10)
                    16) avg BT from 20 to 120 km radius (deg C*10)
                    17)  radius of min BT (km)
                    18-20) Variables need for storm size estimation""",
        "LeadTimeVar": False,
        "Data_Length": 20,
    },
    "IRM1": {
        "long_name": "Predictors from GOES data 1.5hrs before t0",
        "units": "See Notes",
        "Notes": f""" Not time dependent. The 20 values in this record are as follows:
                    1) Time (hhmm) of the GOES image, relative to this case
                    2) Average GOES ch 4 brightness temp (deg C *10), r=0-200 km
                    3) Stan. Dev. of GOES BT (deg C*10), r=0-200 km
                    4) Same as 2) for r=100-300 km
                    5) Same as 3) for r=100-300 km
                    6) Percent area r=50-200 km of GOES ch 4 BT < -10 C
                    7) Same as 6 for BT < -20 C
                    8) Same as 6 for BT < -30 C
                    9) Same as 6 for BT < -40 C
                    10) Same as 6 for BT < -50 C
                    11) Same as 6 for BT < -60 C
                    12) max BT from 0 to 30 km radius (deg C*10)
                    13) avg BT from 0 to 30 km radius (deg C*10)
                    14) radius of max BT (km)
                    15) min BT from 20 to 120 km radius (deg C*10)
                    16) avg BT from 20 to 120 km radius (deg C*10)
                    17)  radius of min BT (km)
                    18-20) Variables need for storm size estimation""",
        "LeadTimeVar": False,
        "Data_Length": 20,
    },
    "IRM3": {
        "long_name": "Predictors from GOES data 3hrs before t0",
        "units": "See Notes",
        "Notes": f""" Not time dependent. The 20 values in this record are as follows:
                    1) Time (hhmm) of the GOES image, relative to this case
                    2) Average GOES ch 4 brightness temp (deg C *10), r=0-200 km
                    3) Stan. Dev. of GOES BT (deg C*10), r=0-200 km
                    4) Same as 2) for r=100-300 km
                    5) Same as 3) for r=100-300 km
                    6) Percent area r=50-200 km of GOES ch 4 BT < -10 C
                    7) Same as 6 for BT < -20 C
                    8) Same as 6 for BT < -30 C
                    9) Same as 6 for BT < -40 C
                    10) Same as 6 for BT < -50 C
                    11) Same as 6 for BT < -60 C
                    12) max BT from 0 to 30 km radius (deg C*10)
                    13) avg BT from 0 to 30 km radius (deg C*10)
                    14) radius of max BT (km)
                    15) min BT from 20 to 120 km radius (deg C*10)
                    16) avg BT from 20 to 120 km radius (deg C*10)
                    17)  radius of min BT (km)
                    18-20) Variables need for storm size estimation""",
        "LeadTimeVar": False,
        "Data_Length": 20,
    },
    "PC00": {
        "long_name": "IR Principal Components and related variables @t0",
        "units": "Unitless",  # confirm with @Mark and @Kate
        "Notes": "",
        "LeadTimeVar": False,
        "Data_Length": 21,  # confirm with @Mark and @Kate
    },
    "PCM1": {
        "long_name": "IR Principal Components and related variables @t0-1.5hrs",
        "units": "Unitless",  # confirm with @Mark and @Kate
        "Notes": "",
        "LeadTimeVar": False,
        "Data_Length": 21,  # confirm with @Mark and @Kate
    },
    "PCM3": {
        "long_name": "IR Principal Components and related variables @t0-3hrs",
        "units": "Unitless",  # confirm with @Mark and @Kate
        "Notes": "",
        "LeadTimeVar": False,
        "Data_Length": 21,  # confirm with @Mark and @Kate
    },
    "RD20": {
        "long_name": "Ocean depth of the 20 deg C isotherm",
        "units": "m",
        "Notes": "from satellite altimetry data",
    },
    "RD26": {
        "long_name": "Ocean depth of the 26 deg C isotherm",
        "units": "m",
        "Notes": "from satellite altimetry data",
    },
    "RHCN": {
        "long_name": "Ocean heat content",
        "units": "KJ/cm2",
        "Notes": "from satellite altimetry data",
        "Info_Label": "Age, in days, of the OHC analysis used to estimate RD20, RD26 and RHCN",
    },
    "NSST": {
        "long_name": "SST from the NCODA analysis",
        "units": "deg C*10",
        "Notes": "",
    },
    "NSTA": {
        "long_name": "Averaged SST from the NCODA analysis",
        "units": "deg C*10",
        "Notes": "spatially averaged over 5 points (storm center, + 50 km N, E, S and W of center)",
    },
    "NTMX": {
        "long_name": "Max ocean temperature in the NCODA vertical profile",
        "units": "deg C*10",
        "Notes": "",
    },
    "NDMX": {
        "long_name": "Depth of the max ocean temperature in the profile",
        "units": "m",
        "Notes": "",
    },
    "NDML": {
        "long_name": "Depth of the mixed layer",
        "units": "m",
        "Notes": """Defined as the depth where the T is 0.5 colder than at the surface. 
                    In rare cases, this is the depth where the T is 0.5 warmer than at the surface.
                    In those cases, NDML is negative.""",  # verify with @Mark and @Kate
    },
    "ND30": {
        "long_name": "Depth of the 30 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND28": {
        "long_name": "Depth of the 28 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND26": {
        "long_name": "Depth of the 26 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND24": {
        "long_name": "Depth of the 24 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND22": {
        "long_name": "Depth of the 22 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND20": {
        "long_name": "Depth of the 20 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND18": {
        "long_name": "Depth of the 18 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "ND16": {
        "long_name": "Depth of the 16 deg C isotherm",
        "units": "m",
        "Notes": "",
    },
    "NDFR": {
        "long_name": "Depth of the lowest model level in the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "NTFR": {
        "long_name": "Ocean T at the lowest level in the NCODA analysis",
        "units": "deg C*10",
        "Notes": "",
    },
    "NOHC": {
        "long_name": "Ocean heat content relative to the 26 C isotherm, from the NCODA analysis",
        "units": "J/kg-deg C",
        "Notes": "",
    },
    "NO20": {
        "long_name": "Ocean heat content relative to the 20 deg C isotherm, from the NCODA analysis",
        "units": "J/kg-deg C",
        "Notes": "",
    },
    "XNST": {
        "long_name": "Climatological values of the SST from the NCODA analysis",
        "units": "deg C*10",
        "Notes": "",
    },
    "XTMX": {
        "long_name": "Climatological values of the max ocean temperature in the NCODA vertical profile",
        "units": "deg C*10",
        "Notes": "",
    },
    "XDMX": {
        "long_name": "Climatological values of the depth of the max ocean temperature in the profile",
        "units": "m",
        "Notes": "",
    },
    "XDML": {
        "long_name": "Climatological values of the depth of the mixed layer",
        "units": "m",
        "Notes": """Defined as the depth where the T is 0.5 colder than at the surface. 
                    In rare cases, this is the depth where the T is 0.5 warmer than at the surface.
                    In those cases, XDML is negative.""",
    },
    "XD16": {
        "long_name": "Climatological values of the depth of the 16 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD18": {
        "long_name": "Climatological values of the depth of the 18 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD20": {
        "long_name": "Climatological values of the depth of the 20 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD22": {
        "long_name": "Climatological values of the depth of the 22 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD24": {
        "long_name": "Climatological values of the depth of the 24 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD26": {
        "long_name": "Climatological values of the depth of the 26 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD28": {
        "long_name": "Climatological values of the depth of the 28 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XD30": {
        "long_name": "Climatological values of the depth of the 30 deg C isotherm from the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XDFR": {
        "long_name": "Climatological values of the depth of the lowest model level in the NCODA analysis",
        "units": "m",
        "Notes": "",
    },
    "XTFR": {
        "long_name": "Climatological values of the Ocean T at the lowest level in the NCODA analysis",
        "units": "deg C*10",
        "Notes": "",
    },
    "XOHC": {
        "long_name": "Climatological values of the Ocean heat content relative to the 26 C isotherm, from the NCODA analysis",
        "units": "KJ/cm2",
        "Notes": "",
    },
    "X020": {
        "long_name": "Climatological values of the Ocean heat content relative to the 20 C isotherm, from the NCODA analysis",
        "units": "J/kg-deg C",
        "Notes": "",
    },
    "XDST": {
        "long_name": "Climatological value of the daily Reynolds SST",
        "units": "deg C*10",
        "Notes": "",
    },
}

# %% Metpy units
[
    "%",
    "A",
    "A_90",
    "A_US",
    "A_it",
    "Ah",
    "At",
    "B",
    "BDFT",
    "BF",
    "BTU",
    "Ba",
    "Bd",
    "Bi",
    "Bq",
    "Btu",
    "Btu_iso",
    "Btu_it",
    "Btu_th",
    "C",
    "C_90",
    "Ci",
    "Cl",
    "Context",
    "D",
    "DPI",
    "Da",
    "ECC",
    "EC_therm",
    "E_h",
    "Eh",
    "F",
    "FBM",
    "F_90",
    "Fr",
    "G",
    "G_0",
    "Gal",
    "Gb",
    "Group",
    "Gy",
    "H",
    "H2O",
    "H_90",
    "Hg",
    "Hg_0C",
    "Hg_32F",
    "Hg_60F",
    "Hz",
    "J",
    "K",
    "KPH",
    "K_J",
    "K_J90",
    "K_alpha_Cu_d_220",
    "K_alpha_Mo_d_220",
    "K_alpha_W_d_220",
    "L",
    "Ly",
    "M",
    "MPH",
    "Measurement",
    "Mx",
    "N",
    "N_A",
    "Ne",
    "NeC",
    "Nm",
    "Np",
    "Oe",
    "P",
    "PPCM",
    "PPI",
    "PSH",
    "Pa",
    "Phi_0",
    "Quantity",
    "R",
    "RKM",
    "R_K",
    "R_K90",
    "R_inf",
    "R_∞",
    "Rd",
    "Ry",
    "S",
    "SPL",
    "St",
    "Sv",
    "System",
    "T",
    "Ta",
    "Td",
    "Tj",
    "Tt",
    "U",
    "UK_bbl",
    "UK_bushel",
    "UK_cup",
    "UK_cwt",
    "UK_fluid_ounce",
    "UK_force_ton",
    "UK_gallon",
    "UK_gill",
    "UK_horsepower",
    "UK_hundredweight",
    "UK_pint",
    "UK_pk",
    "UK_quart",
    "UK_ton",
    "UK_ton_force",
    "US_cwt",
    "US_dry_barrel",
    "US_dry_gallon",
    "US_dry_pint",
    "US_dry_quart",
    "US_fluid_dram",
    "US_fluid_ounce",
    "US_force_ton",
    "US_hundredweight",
    "US_international_ampere",
    "US_international_ohm",
    "US_international_volt",
    "US_liquid_cup",
    "US_liquid_dram",
    "US_liquid_fifth",
    "US_liquid_gallon",
    "US_liquid_gill",
    "US_liquid_ounce",
    "US_liquid_quart",
    "US_pint",
    "US_shot",
    "US_therm",
    "US_ton",
    "US_ton_force",
    "Unit",
    "UnitsContainer",
    "V",
    "VA",
    "V_90",
    "V_US",
    "V_it",
    "W",
    "W_90",
    "Wb",
    "Wh",
    "Xu_Cu",
    "Xu_Mo",
    "Z_0",
    "a",
    "a0",
    "a_0",
    "a_u_action",
    "a_u_current",
    "a_u_electric_field",
    "a_u_energy",
    "a_u_force",
    "a_u_intensity",
    "a_u_length",
    "a_u_mass",
    "a_u_temp",
    "a_u_time",
    "abA",
    "abC",
    "abF",
    "abH",
    "abS",
    "abV",
    "abampere",
    "abcoulomb",
    "aberdeen",
    "abfarad",
    "abhenry",
    "abmho",
    "abohm",
    "absiemens",
    "abvolt",
    "abΩ",
    "acre",
    "acre_feet",
    "acre_foot",
    "add_context",
    "alpha",
    "amp",
    "ampere",
    "ampere_hour",
    "ampere_turn",
    "amu",
    "angstrom",
    "angstrom_star",
    "angular_degree",
    "angular_minute",
    "angular_second",
    "ap_dr",
    "ap_lb",
    "ap_oz",
    "apothecary_dram",
    "apothecary_ounce",
    "apothecary_pound",
    "arc_minute",
    "arc_second",
    "arcdeg",
    "arcdegree",
    "arcmin",
    "arcminute",
    "arcsec",
    "arcsecond",
    "are",
    "astronomical_unit",
    "at",
    "atm",
    "atm_l",
    "atmosphere",
    "atmosphere_liter",
    "atomic_mass_constant",
    "atomic_unit_of_action",
    "atomic_unit_of_current",
    "atomic_unit_of_electric_field",
    "atomic_unit_of_energy",
    "atomic_unit_of_force",
    "atomic_unit_of_intensity",
    "atomic_unit_of_length",
    "atomic_unit_of_mass",
    "atomic_unit_of_temperature",
    "atomic_unit_of_time",
    "au",
    "auto_reduce_dimensions",
    "autoconvert_offset_to_baseunit",
    "autoconvert_to_preferred",
    "avdp_dram",
    "avdp_ounce",
    "avdp_pound",
    "avogadro_constant",
    "avogadro_number",
    "avoirdupois_dram",
    "avoirdupois_ounce",
    "avoirdupois_pound",
    "b",
    "bag",
    "bar",
    "barad",
    "barie",
    "barn",
    "barrel",
    "barrie",
    "baryd",
    "barye",
    "baud",
    "bbl",
    "becquerel",
    "beer_barrel",
    "beer_bbl",
    "big_point",
    "biot",
    "biot_turn",
    "bit",
    "bits_per_pixel",
    "blob",
    "board_feet",
    "board_foot",
    "bohr",
    "bohr_magneton",
    "bohr_radius",
    "boiler_horsepower",
    "boltzmann_constant",
    "bp",
    "bpp",
    "bps",
    "british_thermal_unit",
    "bu",
    "buckingham",
    "bushel",
    "byte",
    "c",
    "c_0",
    "c_1",
    "c_2",
    "cables_length",
    "cache_folder",
    "cal",
    "cal_15",
    "cal_it",
    "cal_th",
    "calorie",
    "candela",
    "candle",
    "carat",
    "case_sensitive",
    "cc",
    "cd",
    "celsius",
    "centimeter",
    "centimeter_H2O",
    "centimeter_Hg",
    "centimeter_Hg_0C",
    "centipoise",
    "centuries",
    "century",
    "chain",
    "characteristic_impedance_of_vacuum",
    "check",
    "cicero",
    "circle",
    "circular_mil",
    "classical_electron_radius",
    "clausius",
    "cmH2O",
    "cmHg",
    "cm_1",
    "cm_H2O",
    "cm_Hg",
    "cmil",
    "common_year",
    "conductance_quantum",
    "context",
    "conventional_ampere_90",
    "conventional_coulomb_90",
    "conventional_farad_90",
    "conventional_henry_90",
    "conventional_josephson_constant",
    "conventional_mercury",
    "conventional_ohm_90",
    "conventional_volt_90",
    "conventional_von_klitzing_constant",
    "conventional_water",
    "conventional_watt_90",
    "convert",
    "cooling_tower_ton",
    "coulomb",
    "coulomb_constant",
    "count",
    "counts_per_second",
    "cp",
    "cps",
    "css_pixel",
    "ct",
    "cu_ft",
    "cu_in",
    "cu_yd",
    "cubic_centimeter",
    "cubic_feet",
    "cubic_foot",
    "cubic_inch",
    "cubic_yard",
    "cup",
    "curie",
    "cwt",
    "cycle",
    "d",
    "dB",
    "dBm",
    "dBu",
    "d_220",
    "dalton",
    "darcy",
    "day",
    "debye",
    "decade",
    "decibel",
    "decibelmicrowatt",
    "decibelmilliwatt",
    "decimeter",
    "decitex",
    "default_as_delta",
    "default_format",
    "default_system",
    "define",
    "deg",
    "degC",
    "degF",
    "degK",
    "degR",
    "degRe",
    "degree",
    "degreeC",
    "degreeE",
    "degreeF",
    "degreeK",
    "degreeN",
    "degreeR",
    "degreeRe",
    "degree_Celsius",
    "degree_E",
    "degree_Fahrenheit",
    "degree_Kelvin",
    "degree_N",
    "degree_Rankine",
    "degree_Reaumur",
    "degree_Réaumur",
    "degree_east",
    "degree_north",
    "degreesE",
    "degreesN",
    "degrees_E",
    "degrees_N",
    "degrees_east",
    "degrees_north",
    "delta_celsius",
    "delta_degC",
    "delta_degF",
    "delta_degRe",
    "delta_degreeC",
    "delta_degreeF",
    "delta_degreeRe",
    "delta_degree_Celsius",
    "delta_degree_Fahrenheit",
    "delta_degree_Reaumur",
    "delta_degree_Réaumur",
    "delta_fahrenheit",
    "delta_reaumur",
    "delta_réaumur",
    "den",
    "denier",
    "dgal",
    "didot",
    "dirac_constant",
    "disable_contexts",
    "dot",
    "dots_per_inch",
    "dpi",
    "dqt",
    "dr",
    "drachm",
    "dram",
    "dry_barrel",
    "dry_gallon",
    "dry_pint",
    "dry_quart",
    "dtex",
    "dwt",
    "dyn",
    "dyne",
    "e",
    "eV",
    "electric_constant",
    "electrical_horsepower",
    "electron_g_factor",
    "electron_mass",
    "electron_volt",
    "elementary_charge",
    "enable_contexts",
    "entropy_unit",
    "enzyme_unit",
    "enzymeunit",
    "eon",
    "eps0",
    "eps_0",
    "epsilon_0",
    "erg",
    "esu",
    "eu",
    "eulers_number",
    "fahrenheit",
    "farad",
    "faraday",
    "faraday_constant",
    "fathom",
    "feet",
    "feet_H2O",
    "femtometer",
    "fermi",
    "fifteen_degree_calorie",
    "fifth",
    "fine_structure_constant",
    "first_radiation_constant",
    "fldr",
    "floz",
    "fluid_dram",
    "fluid_ounce",
    "fluidram",
    "fm",
    "fmt_locale",
    "foot",
    "foot_H2O",
    "foot_per_second",
    "foot_pound",
    "footpound",
    "force_gram",
    "force_kilogram",
    "force_long_ton",
    "force_metric_ton",
    "force_ndarray",
    "force_ndarray_like",
    "force_ounce",
    "force_pound",
    "force_short_ton",
    "force_t",
    "force_ton",
    "fortnight",
    "fps",
    "franklin",
    "ft",
    "ftH2O",
    "ft_lb",
    "fur",
    "furlong",
    "g",
    "g0",
    "g_0",
    "g_e",
    "g_n",
    "gal",
    "galileo",
    "gallon",
    "gamma",
    "gamma_mass",
    "gauss",
    "get_base_units",
    "get_compatible_units",
    "get_dimensionality",
    "get_group",
    "get_name",
    "get_root_units",
    "get_symbol",
    "get_system",
    "gf",
    "gi",
    "gilbert",
    "gill",
    "gon",
    "gpm",
    "gr",
    "grad",
    "grade",
    "grain",
    "gram",
    "gram_force",
    "gravitational_constant",
    "gravity",
    "gray",
    "gregorian_year",
    "h",
    "ha",
    "hand",
    "hartree",
    "hartree_energy",
    "hbar",
    "hectare",
    "hectopascal",
    "henry",
    "hertz",
    "hogshead",
    "horsepower",
    "hour",
    "hp",
    "hr",
    "hundredweight",
    "hydraulic_horsepower",
    "impedance_of_free_space",
    "imperial_barrel",
    "imperial_bbl",
    "imperial_bu",
    "imperial_bushel",
    "imperial_cp",
    "imperial_cup",
    "imperial_fldr",
    "imperial_floz",
    "imperial_fluid_drachm",
    "imperial_fluid_dram",
    "imperial_fluid_ounce",
    "imperial_fluid_scruple",
    "imperial_gal",
    "imperial_gallon",
    "imperial_gi",
    "imperial_gill",
    "imperial_minim",
    "imperial_peck",
    "imperial_pint",
    "imperial_pk",
    "imperial_pt",
    "imperial_qt",
    "imperial_quart",
    "in",
    "inHg",
    "in_Hg",
    "inch",
    "inch_H2O_39F",
    "inch_H2O_60F",
    "inch_Hg",
    "inch_Hg_32F",
    "inch_Hg_60F",
    "inches",
    "international_british_thermal_unit",
    "international_calorie",
    "international_feet",
    "international_foot",
    "international_inch",
    "international_inches",
    "international_knot",
    "international_mile",
    "international_steam_table_calorie",
    "international_yard",
    "is_compatible_with",
    "jig",
    "josephson_constant",
    "joule",
    "julian_year",
    "jute",
    "k",
    "k_B",
    "k_C",
    "karat",
    "kat",
    "katal",
    "kayser",
    "kelvin",
    "kgf",
    "kilogram",
    "kilogram_force",
    "kilojoule",
    "kilometer",
    "kilometer_per_hour",
    "kilometer_per_second",
    "kip",
    "kip_per_square_inch",
    "knot",
    "knot_international",
    "kph",
    "kps",
    "ksi",
    "kt",
    "l",
    "lambda",
    "lambert",
    "langley",
    "lattice_spacing_of_Si",
    "lb",
    "lbf",
    "lbt",
    "league",
    "leap_year",
    "li",
    "light_year",
    "lightyear",
    "link",
    "liquid_cup",
    "liquid_gallon",
    "liquid_gill",
    "liquid_pint",
    "liquid_quart",
    "liter",
    "litre",
    "lm",
    "ln10",
    "load_definitions",
    "long_hundredweight",
    "long_ton",
    "long_ton_force",
    "lumen",
    "lunar_month",
    "lux",
    "lx",
    "ly",
    "m",
    "m_e",
    "m_n",
    "m_p",
    "m_u",
    "magnetic_constant",
    "magnetic_flux_quantum",
    "mas",
    "maxwell",
    "mean_international_ampere",
    "mean_international_ohm",
    "mean_international_volt",
    "mercury",
    "mercury_60F",
    "meter",
    "meter_per_second",
    "metre",
    "metric_horsepower",
    "metric_ton",
    "metric_ton_force",
    "mho",
    "mi",
    "microgram",
    "microliter",
    "micrometer",
    "micromole",
    "micron",
    "mil",
    "mil_length",
    "mile",
    "mile_per_hour",
    "millennia",
    "millennium",
    "milliarcsecond",
    "millibar",
    "milligram",
    "millimeter",
    "millimeter_Hg",
    "millimeter_Hg_0C",
    "min",
    "minim",
    "minute",
    "mmHg",
    "mm_Hg",
    "mol",
    "molar",
    "molar_gas_constant",
    "mole",
    "molec",
    "molecule",
    "month",
    "mph",
    "mpl_formatter",
    "mps",
    "mu0",
    "mu_0",
    "mu_B",
    "mu_N",
    "nautical_mile",
    "neper",
    "neutron_mass",
    "newton",
    "newtonian_constant_of_gravitation",
    "nit",
    "nmi",
    "non_int_type",
    "nuclear_magneton",
    "number_english",
    "number_meter",
    "oct",
    "octave",
    "octet",
    "oersted",
    "ohm",
    "ohm_90",
    "ohm_US",
    "ohm_it",
    "oil_barrel",
    "oil_bbl",
    "ounce",
    "ounce_force",
    "oz",
    "ozf",
    "ozt",
    "params",
    "parse_expression",
    "parse_pattern",
    "parse_unit_name",
    "parse_units",
    "parse_units_as_container",
    "parsec",
    "particle",
    "pascal",
    "pc",
    "pdl",
    "peak_sun_hour",
    "peck",
    "pel",
    "pennyweight",
    "percent",
    "perch",
    "pi",
    "pi_theorem",
    "pica",
    "picture_element",
    "pint",
    "pixel",
    "pixels_per_centimeter",
    "pixels_per_inch",
    "pk",
    "planck_constant",
    "planck_current",
    "planck_length",
    "planck_mass",
    "planck_temperature",
    "planck_time",
    "point",
    "poise",
    "pole",
    "pond",
    "pound",
    "pound_force",
    "pound_force_per_square_inch",
    "poundal",
    "pp",
    "ppi",
    "ppm",
    "preprocessors",
    "printers_dpi",
    "printers_pica",
    "printers_point",
    "proton_mass",
    "psi",
    "pt",
    "px",
    "qt",
    "quad",
    "quadrillion_Btu",
    "quart",
    "quarter",
    "r_e",
    "rad",
    "radian",
    "rads",
    "rankine",
    "rd",
    "reaumur",
    "reciprocal_centimeter",
    "refrigeration_ton",
    "rem",
    "remove_context",
    "revolution",
    "revolutions_per_minute",
    "revolutions_per_second",
    "reyn",
    "rhe",
    "rod",
    "roentgen",
    "rpm",
    "rps",
    "rutherford",
    "rydberg",
    "rydberg_constant",
    "réaumur",
    "röntgen",
    "s",
    "scaled_point",
    "scruple",
    "sec",
    "second",
    "second_radiation_constant",
    "section",
    "separate_format_defaults",
    "set_fmt_locale",
    "setup_matplotlib",
    "sft",
    "shake",
    "short_hundredweight",
    "short_ton",
    "short_ton_force",
    "shot",
    "sidereal_day",
    "sidereal_month",
    "sidereal_year",
    "siemens",
    "sievert",
    "sigma",
    "sigma_e",
    "slinch",
    "slm",
    "slpm",
    "slug",
    "slugette",
    "smi",
    "sound_pressure_level",
    "speed_of_light",
    "sq_deg",
    "sq_ft",
    "sq_in",
    "sq_mi",
    "sq_perch",
    "sq_pole",
    "sq_rod",
    "sq_yd",
    "sqdeg",
    "square_degree",
    "square_feet",
    "square_foot",
    "square_inch",
    "square_inches",
    "square_league",
    "square_mile",
    "square_rod",
    "square_survey_mile",
    "square_yard",
    "sr",
    "standard_atmosphere",
    "standard_gravity",
    "standard_liter_per_minute",
    "statA",
    "statC",
    "statF",
    "statH",
    "statT",
    "statV",
    "statWb",
    "statampere",
    "statcoulomb",
    "statfarad",
    "stathenry",
    "statmho",
    "statohm",
    "stattesla",
    "statvolt",
    "statweber",
    "statΩ",
    "stefan_boltzmann_constant",
    "steradian",
    "stere",
    "stilb",
    "stokes",
    "stone",
    "super_feet",
    "super_foot",
    "superficial_feet",
    "superficial_foot",
    "survey_foot",
    "survey_link",
    "survey_mile",
    "sv",
    "svedberg",
    "sverdrup",
    "synodic_month",
    "sys",
    "t",
    "tTNT",
    "t_force",
    "tablespoon",
    "tansec",
    "tbsp",
    "teaspoon",
    "technical_atmosphere",
    "tesla",
    "tex",
    "tex_cicero",
    "tex_didot",
    "tex_pica",
    "tex_point",
    "tf",
    "th",
    "therm",
    "thermochemical_british_thermal_unit",
    "thermochemical_calorie",
    "thm",
    "thomson_cross_section",
    "thou",
    "tlb",
    "toe",
    "ton",
    "ton_TNT",
    "ton_force",
    "ton_of_refrigeration",
    "tonne",
    "tonne_of_oil_equivalent",
    "torr",
    "townsend",
    "toz",
    "tropical_month",
    "tropical_year",
    "troy_ounce",
    "troy_pound",
    "tsp",
    "turn",
    "u",
    "unified_atomic_mass_unit",
    "unit_pole",
    "us_statute_mile",
    "vacuum_permeability",
    "vacuum_permittivity",
    "volt",
    "volt_ampere",
    "von_klitzing_constant",
    "water",
    "water_39F",
    "water_4C",
    "water_60F",
    "watt",
    "watt_hour",
    "watthour",
    "weber",
    "week",
    "wien_frequency_displacement_law_constant",
    "wien_u",
    "wien_wavelength_displacement_law_constant",
    "wien_x",
    "with_context",
    "wraps",
    "x_unit_Cu",
    "x_unit_Mo",
    "yard",
    "yd",
    "year",
    "yr",
    "zeta",
    "°C",
    "°F",
    "°K",
    "°R",
    "°Re",
    "µ",
    "µ_0",
    "µ_B",
    "µ_N",
    "Å",
    "Å_star",
    "ångström",
    "ørsted",
    "ħ",
    "Δcelsius",
    "ΔdegC",
    "ΔdegF",
    "ΔdegRe",
    "ΔdegreeC",
    "ΔdegreeF",
    "ΔdegreeRe",
    "Δdegree_Réaumur",
    "Δfahrenheit",
    "Δreaumur",
    "Δréaumur",
    "Δ°C",
    "Δ°F",
    "Δ°Re",
    "Φ_0",
    "Ω",
    "Ω_90",
    "Ω_US",
    "Ω_it",
    "α",
    "γ",
    "ε_0",
    "ζ",
    "λ",
    "μ",
    "π",
    "σ",
    "σ_e",
    "ℎ",
    "Å",
]



# Climatology Baseline Models
# Date: 2024.05.24
# This file contains the implementation of baseline models that output climatology.
# Author: Milton Gomez
# %% Imports
# OS and IO
import os
import sys
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import dask.array as da
import pandas as pd


# ML Libraries
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.base import BaseEstimator
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import SGDRegressor, LinearRegression
from sklearn.pipeline import make_pipeline

# Backend Libraries
import joblib as jl

from ..utils import toolbox, constants, ML_functions as mlf
from ..utils import data_lib as dlib


# %% Utility Functions
def PolynomialRegression(degree=3, **kwargs):
    return make_pipeline(PolynomialFeatures(degree), LinearRegression(**kwargs))


def exponential_regressor(x, y):
    # convert to and ln(y)
    y = np.log(y)

    # Fit a linear regression model
    model = LinearRegression()
    model.fit(x[:, np.newaxis], y)

    # Get the parameters of the linear regression model
    slope = model.coef_[0]
    intercept = model.intercept_

    return lambda x: np.exp(intercept) * np.exp(slope * x)


def nearest_interpolator(values, df):
    # Get the first and second columns of the DataFrame
    first_col = df.iloc[:, 0].to_numpy()
    second_col = df.iloc[:, 1].to_numpy()

    # Initialize an empty list to store the results
    results = []

    # For each value in the input array, find the nearest value in the first column
    for value in values:
        diff = np.abs(first_col - value)
        idx = np.argmin(diff)
        nearest_value = second_col[idx]
        results.append(nearest_value)

    return np.array(results)


def plot_facecolors(fig, axes, **kwargs):
    figcolor = kwargs.get("figcolor", np.array([1, 21, 38]) / 255)
    axcolor = kwargs.get("axcolor", np.array([141, 166, 166]) / 255)
    textcolor = kwargs.get("textcolor", np.array([242, 240, 228]) / 255)

    # Change the facecolor of the figure
    fig.set_facecolor(figcolor)

    # Change the facecolor of the axes
    for ax in axes.flatten():
        ax.set_facecolor(axcolor)

    # Change the color of the title
    for ax in axes.flatten():
        title = ax.get_title()
        ax.set_title(title, color=textcolor)

    # Change the color of the x and y axis labels
    for ax in axes.flatten():
        ax.xaxis.label.set_color(textcolor)
        ax.yaxis.label.set_color(textcolor)

    # Change the color of the tick labels
    for ax in axes.flatten():
        for label in ax.get_xticklabels():
            label.set_color(textcolor)
        for label in ax.get_yticklabels():
            label.set_color(textcolor)


# %%
class DeltaIntensityClimatology(BaseEstimator):
    """
    Intensity climatology baseline model.
    This model predicts the intensity climatology based on the maximum wind speed
    in the training data. In order to calculate this, the model calculates the mean
    and standard deviation of the intensification rate between time steps in the
    training set of storms.
    """

    def __repr__(self):
        return f"TCBench Climatology Baseline - {self.name}"

    def __init__(self):
        self.name = "Delta Intensity Climatology"

    def fit(self, X, y=None):
        # Assert that the training set is a dictionary with numeric keys
        # that each point to a list of storm objects
        assert isinstance(X, dict), "Training data must be a dictionary."
        # assert all(
        #     [isinstance(key, int) for key in X.keys()]
        # ), "Keys of the training data must be integers."
        assert all(
            [isinstance(value, list) for value in X.values()]
        ), "Values of the training data must be lists."
        assert all(
            [
                all([isinstance(storm, toolbox.tc_track) for storm in value])
                for value in X.values()
            ]
        ), "Values of the training data must be lists of TCBench TC Tracks."

        # Concatenate the list of storms into a single list
        storm_list = []
        for key in X.keys():
            storm_list += X[key]

        records = []
        for storm in storm_list:
            # Make a boolean index for the values of storm.wind that are a numeric string
            numeric_bool = np.char.isnumeric(storm.wind.astype(str))

            # Make a boolean masks for the timestamps in 0, 6, 12, and 18 h
            time_bool = np.isin(pd.to_datetime(storm.timestamps).hour, [0, 6, 12, 18])

            # Copy the wind
            wind = storm.wind.copy()
            wind[~numeric_bool] = np.NaN
            wind = wind.astype(float)
            wind = wind[time_bool]

            # Calculate the intensification rate
            base_intensity = wind[:-1]
            intensification_rate = np.diff(wind)

            # Filter out the NaN values
            intensification_rate = intensification_rate[~np.isnan(base_intensity)]
            base_intensity = base_intensity[~np.isnan(base_intensity)]

            base_intensity = base_intensity[~np.isnan(intensification_rate)]
            intensification_rate = intensification_rate[~np.isnan(intensification_rate)]

            records.append(np.vstack([base_intensity, intensification_rate]).T)

        # Concatenate the records
        records = np.vstack(records)

        self.intensity_climatology = None
        unique_intensities = np.unique(records[:, 0])

        self.training_data = records

        # intensity_distribution = pd.DataFrame(
        #     np.vstack([unique_intensities, np.zeros(unique_intensities.shape[0])]).T,
        #     columns=["Intensity", "Count"],
        # )

        for intensity in unique_intensities:
            # Calculate the mean and standard deviation of the intensification rate
            mean = np.mean(records[records[:, 0] == intensity, 1])
            std = np.std(records[records[:, 0] == intensity, 1])

            # intensity_distribution.loc[intensity] += records[
            #     records[:, 0] == intensity
            # ].shape[0]

            # Save the mean and standard deviation using a pandas dataframe
            temp_df = pd.DataFrame(
                [[intensity, mean, std]],
                columns=[
                    "Intensity",
                    "Mean Intensification Rate",
                    "Sigma Intensification Rate",
                ],
            )
            if self.intensity_climatology is None:
                self.intensity_climatology = temp_df
            else:
                self.intensity_climatology = pd.concat(
                    [
                        self.intensity_climatology,
                        temp_df,
                    ]
                )

        self.mu_interpolator = lambda x: nearest_interpolator(
            x, self.intensity_climatology[["Intensity", "Mean Intensification Rate"]]
        )

        # find the intensities and sigma intensification rates fir which the
        # sigma is greater than 2.5, half of the "normal" bin size for intensity
        # reporting
        sigma_mask = self.intensity_climatology["Sigma Intensification Rate"] > 2.5

        # make a 3rd degree polynomial regression model for the sigma intensification rate
        self.sigma_regressor = PolynomialRegression(degree=4)
        self.sigma_regressor.fit(
            self.intensity_climatology["Intensity"][sigma_mask]
            .to_numpy()
            .reshape(-1, 1),
            self.intensity_climatology["Sigma Intensification Rate"][
                sigma_mask
            ].to_numpy(),
        )

        # self.sigma_regressor = exponential_regressor(
        #     self.intensity_climatology["Intensity"][sigma_mask].to_numpy(),
        #     self.intensity_climatology["Sigma Intensification Rate"][
        #         sigma_mask
        #     ].to_numpy(),
        # )

    def summary(self, **kwargs):
        assert self.intensity_climatology is not None, "Model has not been trained."
        assert (
            self.training_data is not None
        ), "Model training record not found - did you use the fit method?"

        fig, axes = plt.subplots(2, 2, figsize=(12, 12), dpi=200)
        int_bins = np.arange(
            self.training_data[:, 0].min(), self.training_data[:, 0].max(), 5
        )
        rate_bins = (
            np.arange(self.training_data[:, 1].min(), self.training_data[:, 1].max(), 5)
            - 2.5
        )
        axes[0, 0].hist(
            self.training_data[:, 0],
            bins=int_bins,  # bins=np.unique(self.training_data[:, 0])
            density=True,
        )
        axes[0, 0].set_title("Intensity Distribution")
        axes[0, 0].set_xlabel("Intensity (kt)")
        axes[0, 0].set_ylabel("Density")

        # ------------------------------------------------
        axes[0, 1].scatter(
            self.training_data[:, 0], self.training_data[:, 1], alpha=0.1
        )
        axes[0, 1].set_title("Intensification Rate vs. Intensity")
        axes[0, 1].set_xlabel("Intensity (kt)")
        axes[0, 1].set_ylabel("6h Intensification Rate (kt)")

        axes[1, 0].hist(
            self.training_data[:, 1],
            bins=rate_bins,  # bins=np.unique(self.training_data[:, 1])
            density=True,
        )
        axes[1, 0].set_title("Intensification Rate Distribution")
        axes[1, 0].set_xlabel("6h Intensification Rate (kt)")
        axes[1, 0].set_ylabel("Density")

        # ------------------------------------------------
        axes[1, 1].hist2d(
            self.training_data[:, 0],
            self.training_data[:, 1],
            bins=[
                int_bins,
                rate_bins,
            ],
            density=True,
            cmap="Greens",
        )
        axes[1, 1].set_title("Joint Distribution")
        axes[1, 1].set_xlabel("Intensity (kt)")
        axes[1, 1].set_ylabel("6h Intensification Rate (kt)")

        # Colorize the plot
        plot_facecolors(fig, axes)

        # Show figure
        plt.tight_layout()
        plt.show()
        plt.close()

        # ------------------------------------------------

        fig, axes = plt.subplots(2, 1, figsize=(12, 6), dpi=200)

        # Plot the intensity vs mean intensification rate
        axes[0].scatter(
            self.intensity_climatology["Intensity"],
            self.intensity_climatology["Mean Intensification Rate"],
            label="Mean Intensification Rate",
        )
        axes[0].set_title("Intensity vs. Mean Intensification Rate")
        axes[0].set_xlabel("Intensity")
        axes[0].set_ylabel("Mean Intensification Rate")

        axlineopts = kwargs.get(
            "axlineopts", {"color": "black", "linewidth": 2, "linestyle": "--"}
        )

        # draw the horizontal axis
        axes[0].axhline(0, **axlineopts)
        axes[0].legend()

        # Plot the intensity vs sigma intensification rate
        axes[1].scatter(
            self.intensity_climatology["Intensity"],
            self.intensity_climatology["Sigma Intensification Rate"],
            label="Sigma Intensification Rate",
        )

        xmin = self.intensity_climatology["Intensity"].min()
        xmax = self.intensity_climatology["Intensity"].max() + 10
        x = np.linspace(xmin, xmax, 300).reshape(-1, 1)

        # Plot the sigma intensification rate regressor
        axes[1].plot(
            x,
            self.sigma_regressor.predict(x),
            label="Sigma Intensification Rate Regressor",
        )

        axes[1].set_title("Intensity vs. Sigma Intensification Rate")
        axes[1].set_xlabel("Intensity")
        axes[1].set_ylabel("Sigma Intensification Rate")
        axes[1].axhline(0, **axlineopts)
        axes[1].legend()

        # Colorize the plot
        plot_facecolors(fig, axes)

        # Show figure
        plt.tight_layout()
        plt.show()
        plt.close()

        # ------------------------------------------------

        # plot uncertainty regions
        fig, ax = plt.subplots(1, 1, figsize=(6, 3), dpi=200)

        x = np.arange(10, 190, 5).reshape(-1, 1)
        data = self.predict(x)

        ax.plot(x, data[:, 0], label="Mean Intensification Rate")
        ax.fill_between(
            x.flatten(),
            data[:, 0] - data[:, 1],
            data[:, 0] + data[:, 1],
            alpha=0.5,
            label="Uncertainty Region",
            color="orange",
        )
        ax.set_title("Prediction Sweep (10-190 kts)")
        ax.set_xlabel("Intensity")
        ax.set_ylabel("Intensification Rate")

        # Colorize the plot
        plot_facecolors(fig, np.array([ax]))

        # Show figure
        plt.tight_layout()
        plt.show()
        plt.close()

    def predict(self, X):
        """
        Predict the intensity climatology.
        """
        assert self.intensity_climatology is not None, "Model has not been trained."

        # Get the unique intensities in the input data
        mu = self.mu_interpolator(X)
        sigma = self.sigma_regressor.predict(X)

        return np.vstack([mu, sigma]).T

    def storm_prediction_plotter(self, storm, **kwargs):

        stormline_color = kwargs.get("stormline_color", np.array([242, 92, 5]) / 255)
        tick_color = kwargs.get("tick_color", np.array([242, 240, 228]) / 255)

        wind = storm.wind.astype(str)

        times = storm.timestamps[np.char.isnumeric(wind)]
        wind = wind[np.char.isnumeric(wind)].astype(float)

        wind = wind[np.isin(pd.to_datetime(times).hour, [0, 6, 12, 18])]
        times = times[np.isin(pd.to_datetime(times).hour, [0, 6, 12, 18])]

        rates = self.predict(wind.reshape(-1, 1))

        predicted_winds = wind + rates[:, 0]
        predicted_times = times + np.timedelta64(6, "h")

        fig, ax = plt.subplots(1, 1, figsize=(6, 6), dpi=200)

        ax.scatter(
            predicted_times, predicted_winds, label="Predictions", color=stormline_color
        )
        ax.fill_between(
            predicted_times.flatten(),
            predicted_winds - rates[:, 1],
            predicted_winds + rates[:, 1],
            alpha=0.5,
            label="Uncertainty Region",
            color="orange",
        )
        ax.set_title(
            f"6h Climatological Prediction for Storm {storm.uid} - {storm.name}"
        )
        ax.set_xlabel("Time")
        ax.set_ylabel("Intensity")
        ax.legend()
        ax.scatter(times, wind, label="Ground Truth Intensity", color="black")

        import matplotlib.dates as mdates

        # Set minor ticks to every 6 hours
        hours = mdates.HourLocator(interval=6)
        ax.xaxis.set_minor_locator(hours)

        # Set x-axis tick label color
        ax.tick_params(axis="both", colors=tick_color)
        # ax.xaxis.set_minor_locator(AutoMinorLocator())

        # Enable grid + color
        ax.grid(True)
        ax.grid(color=tick_color)
        # ax.grid(which="minor", color=tick_color, alpha=0.5)

        plt.xticks(rotation=45)

        # Colorize the plot
        plot_facecolors(fig, np.array([ax]))

        # Show figure
        plt.tight_layout()
        plt.show()
        plt.close()


class TC_DeltaIntensity_MLR(BaseEstimator):
    """
    Intensity Linear model baseline.
    This model predicts the change in intensity of a storm based on the intensity of
    the previous time step + the max wind speed predicted by the AI model
    """

    def __repr__(self):
        return f"TCBench Climatology Baseline - {self.name}"

    def __init__(self, **kwargs):
        self.name = "Delta Intensity MLR"
        self.model_type = LinearRegression
        self.u_v_indices = kwargs.get("u_v_indices", [0, 1])
        self.msl_index = kwargs.get("msl_index", 2)
        self.lr = kwargs.get("lr", 0.01)

    def prepare_data(self, base_intensity, AI_X):
        wind_magnitude = np.sqrt(
            AI_X[:, self.u_v_indices[0]] ** 2 + AI_X[:, self.u_v_indices[1]] ** 2
        )
        wind_magnitude = wind_magnitude.max(axis=[*range(1, wind_magnitude.ndim)])
        wind_magnitude = wind_magnitude.reshape(-1, 1)
        mslp = AI_X[:, self.msl_index]
        mslp = mslp.max(axis=[*range(1, mslp.ndim)])
        mslp = mslp.reshape(-1, 1)

        assert (
            base_intensity.shape[0] == AI_X.shape[0]
        ), "Unequal number of samples. Check AI model input size vs intensity size."

        assert (
            base_intensity.shape[1] == 2
        ), "Intensity array must have shape (n_samples, 2) corresponding to wind (idx_0) and pressure (idx_1)."

        return np.hstack([wind_magnitude, mslp, base_intensity])

    def fit(self, base_intensity, AI_X, y=None):
        X = self.prepare_data(base_intensity, AI_X)
        X = X.rechunk({1: X.shape[0] // 1000}).compute()
        delta = (y - base_intensity).rechunk({0: y.shape[0] // 1000}).compute()

        if len(y.shape) == 1:
            y = y.reshape(-1, 1)

        self.models = []
        for i in range(y.shape[1]):
            model = self.model_type()
            model.fit(X, delta[:, i])
            self.models.append(model)

    def predict(self, base_intensity, AI_X):
        X = self.prepare_data(base_intensity, AI_X)

        y_preds = []
        for i in range(len(self.models)):
            y_preds.append(self.models[i].predict(X))

        return np.vstack(y_preds).T + base_intensity

    pass


class TC_DeltaIntensity_CNN(nn.Module):
    def __str__(self):
        return "Dilated_CNN"

    def __init__(self, **kwargs):
        super(TC_DeltaIntensity_CNN, self).__init__()

        # Assume inputs have 241x241 dimensions in channels (u, v, mslp, t_850, z_500)

        conv_depths = kwargs.get("depths", [32, 16, 16, 64, 96])

        # Layer that sees all inputs
        # Output size = 120x120
        self.conv1 = nn.Conv2d(5, conv_depths[0], kernel_size=(2, 2), padding=0)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # First context layer
        # Output size = 120x120
        self.conv2 = nn.Conv2d(
            5, conv_depths[1], kernel_size=(3, 3), padding=1, stride=2, dilation=2
        )

        # Second context layer
        # Output size = 60x60
        # self.conv3 = nn.Conv2d(
        #     5, conv_depths[2], kernel_size=(3, 3), padding=1, stride=4, dilation=3
        # )
        self.conv3 = nn.Conv2d(
            5, conv_depths[2], kernel_size=(3, 3), padding=4, stride=4, dilation=6
        )

        # layer to convolve the first output with the first context layer.
        # There are 16+32=48 input channels and 64 output channels
        # The output dimensions are 120x120
        self.conv4 = nn.Conv2d(48, conv_depths[3], kernel_size=(3, 3), padding=1)

        # layer to convolve the second output with the second context layer.
        # There are 64 + 16 = 80 input channels and 96 output channels
        # The output dimensions are 60x60
        self.conv5 = nn.Conv2d(80, conv_depths[4], kernel_size=(3, 3), padding=1)

        # dense layer to output the prediction. We predict the mean and standard
        # deviation of the intensity change, both for wind and pressure
        self.fc1 = nn.Linear(conv_depths[4] * 60 * 60, 128)

        # We will encode the baseline intensity with a dense layer
        self.fc2 = nn.Linear(2, 16)

        # And make the prediction from the two dense layers
        self.fc3 = nn.Linear(128 + 16, 2 if kwargs.get("deterministic", False) else 4)

    def forward(self, x, base_int):
        # Get the context from the inputs
        x_context1 = F.relu(self.conv2(x))
        x_context2 = F.relu(self.conv3(x))

        # Apply first convolutional layer
        x = self.pool(F.relu(self.conv1(x)))

        # Apply the first context layer
        x = torch.cat([x, x_context1], dim=1)
        x = F.relu(self.conv4(x))
        x = self.pool(x)

        # Apply the second context layer
        x = torch.cat([x, x_context2], dim=1)
        x = F.relu(self.conv5(x))

        # Flatten the output
        x = x.view(-1, 60 * 60 * 96)

        # Apply the dense layers
        x = F.relu(self.fc1(x))
        base_int = torch.squeeze(F.relu(self.fc2(base_int)))

        # Concatenate the base intensity with the output of the dense layer
        x = torch.cat([x, base_int], dim=1)

        x = self.fc3(x)

        return x


class Regularized_Dilated_CNN(TC_DeltaIntensity_CNN):
    def __str__(self):
        return "Regularized_Dilated_CNN"

    def __init__(self, **kwargs):
        super(Regularized_Dilated_CNN, self).__init__(**kwargs)
        self.dropout2d = nn.Dropout2d(kwargs.get("dropout2d", 0.25))
        self.dropout = nn.Dropout(kwargs.get("dropout", 0.25))

    def forward(self, x, base_int):
        # Get the context from the inputs
        x_context1 = F.relu(self.conv2(x))
        x_context2 = F.relu(self.conv3(x))

        # Apply first convolutional layer
        x = self.pool(F.relu(self.conv1(x)))

        # Apply 2d dropout
        x = self.dropout2d(x)

        # Apply the first context layer
        x = torch.cat([x, x_context1], dim=1)
        x = F.relu(self.conv4(x))
        x = self.pool(x)

        # Apply 2d dropout
        x = self.dropout2d(x)

        # Apply the second context layer
        x = torch.cat([x, x_context2], dim=1)
        x = F.relu(self.conv5(x))

        # Apply 2d dropout
        x = self.dropout2d(x)

        # Flatten the output
        x = x.view(-1, 60 * 60 * 96)

        # Apply the dense layers
        x = F.relu(self.fc1(x))
        x = self.dropout(x)
        base_int = torch.squeeze(F.relu(self.fc2(base_int)))

        # Concatenate the base intensity with the output of the dense layer
        x = torch.cat([x, base_int], dim=1)
        # and dropout
        x = self.dropout(x)

        x = self.fc3(x)

        return x


class LinearRegressor(nn.Module):
    def __init__(self, input_dim):
        super(LinearRegressor, self).__init__()
        self.linear = nn.Linear(input_dim, 1)  # Linear layer

    def forward(self, x):
        return self.linear(x)


class SimpleCNN(nn.Module):
    def __str__(self):
        return "Simple_CNN"

    def __init__(self, **kwargs):
        super(SimpleCNN, self).__init__()
        self.conv1 = nn.Conv2d(5, 32, kernel_size=3, stride=1, padding=1)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.conv3 = nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1)
        self.fc1 = nn.Linear(128 * 30 * 30, 512)
        # We will encode the baseline intensity with a dense layer
        self.fc2 = nn.Linear(2, 16)
        self.fc3 = nn.Linear(512 + 16, 2 if kwargs.get("deterministic", False) else 4)

    def forward(self, x, base_int):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = self.pool(F.relu(self.conv3(x)))
        x = x.view(-1, 128 * 30 * 30)
        x = F.relu(self.fc1(x))
        base_int = torch.squeeze(F.relu(self.fc2(base_int)))
        # Concatenate the base intensity with the output of the dense layer
        x = torch.cat([x, base_int], dim=1)
        x = self.fc3(x)
        return x


# %%



# %% Imports
# OS and IO
import os
import sys
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import dask.array as da
from dask.ml import preprocessing

# Backend Libraries
import joblib as jl

from utils import toolbox, ML_functions as mlf
from utils import data_lib as dlib
import metrics.metrics as metrics

# %%
datadir = "/work/FAC/FGSE/IDYST/tbeucler/default/raw_data/TCBench_alpha"

# %% Data Loading
rng = np.random.default_rng(seed=42)

years = list(range(2013, 2020))
rng.shuffle(years)

sets, data = toolbox.get_sets(
    {"train": years[:-2], "test": [2020], "validation": years[-2:]},
    datadir=datadir,
    test_strategy="custom",
    # use_cached=False,
    # verbose=True,
    # debug=True,
)

# %% Reload for interactive development
try:
    import importlib

    importlib.reload(TCBench.baselines.baselines)
    importlib.reload(metrics)
except:
    import tcbenchmark.baselines.baselines as baselines

# %% MLR Training

# We begin by instantiating our baseline model
MLR = baselines.TC_DeltaIntensity_MLR()

# And fitting it using the training data. Note that we don't
# normalize the inputs
MLR.fit(
    base_intensity=data["train"]["base_intensity"],
    AI_X=data["train"]["inputs"],
    y=data["train"]["outputs"],
)
# %% MLR Validation

# We then use the model to predict the validation data
y_hat = MLR.predict(
    base_intensity=data["validation"]["base_intensity"],
    AI_X=data["validation"]["inputs"],
)

# %% Baseline

# We want to compare the model to a simple baseline, in this case
# the persistence model. Note that this is simply the base intensity
# we're predicting the delta for.
y_persistence = data["validation"]["base_intensity"]

# %% Evaluation

# Let's start by loading the validation outputs into a variable
# for easier access
y_true = data["validation"]["outputs"]

# %% Compute y_true, y_hat, y_persistence
y_true = y_true.compute()
y_hat = y_hat.compute()
y_persistence = y_persistence.compute()

# %% Evaluation
# We will want to evaluate the model using the Root Mean Squared Error and
# the Mean Absolute Error, as well as the associated skill scores compared
# to the persistence model

# Importing the metrics
from sklearn.metrics import root_mean_squared_error, mean_absolute_error

global_performance = metrics.summarize_performance(
    y_true,  # Ground truth
    y_hat,  # Model prediction
    y_persistence,  # Baseline, used for skill score calculations
    [root_mean_squared_error, mean_absolute_error],
)
# %%
# Plotting Global Performance
fig, axes = plt.subplots(
    1, 2, figsize=(15, 5), dpi=150, gridspec_kw={"width_ratios": [2, 1]}
)
metrics.plot_performance(
    global_performance, axes, model_name="MLR", baseline_name="Persistence"
)
toolbox.plot_facecolors(fig=fig, axes=axes)
plt.show()
plt.close()

# %%
# Let's also go ahead and do the analysis per lead time
unique_leads = np.unique(data["validation"]["leadtime"].compute())
num_leads = len(unique_leads)

# set up the axes for the lead time plots

fig, axes = plt.subplots(
    num_leads,  # One row per lead time
    2,
    figsize=(15, 5 * num_leads),
    dpi=150,
    gridspec_kw={"width_ratios": [2, 1]},
)
for idx, lead in enumerate(unique_leads):
    lead_mask = data["validation"]["leadtime"] == lead

    y_true_lead = y_true[lead_mask]
    y_hat_lead = y_hat[lead_mask]
    y_persistence_lead = y_persistence[lead_mask]

    lead_performance = metrics.summarize_performance(
        y_true_lead,  # Ground truth
        y_hat_lead,  # Model prediction
        y_persistence_lead,  # Baseline, used for skill score calculations
        [root_mean_squared_error, mean_absolute_error],
    )

    metrics.plot_performance(
        lead_performance, axes[idx], model_name="MLR", baseline_name="Persistence"
    )
    toolbox.plot_facecolors(fig=fig, axes=axes[idx])

    # Append the lead time to the title for both axes
    axes[idx][0].set_title(f"\n Lead Time: +{lead}h \n" + axes[idx][0].get_title())
    axes[idx][1].set_title(f"\n Lead Time: +{lead}h \n" + axes[idx][1].get_title())

# %%



# Climatology Baseline Models
# Date: 2024.05.24
# This file contains the metrics used for evaluating model performance.
# The metrics are implemented following the scikit-learn convention.
# Author: Milton Gomez

# From SKLearn:
# Functions named as ``*_score`` return a scalar value to maximize: the higher
# the better.

# Function named as ``*_error`` or ``*_loss`` return a scalar value to minimize:
# the lower the better.

# %% Imports
# OS and IO
import os
import sys
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import dask.array as da
import pandas as pd

try:
    import torch
except:
    print("torch not available")

# Backend Libraries
import joblib as jl

from ..utils import toolbox, constants, ML_functions as mlf
from ..utils import data_lib as dlib

# %% Reference Dictionaries
short_names = {
    "root_mean_squared_error": "RMSE",
    "mean_squared_error": "MSE",
    "mean_absolute_error": "MAE",
    "mean_absolute_percentage_error": "MAPE",
    "mean_squared_logarithmic_error": "MSLE",
    "r2_score": "R2",
    "explained_variance_score": "EV",
    "max_error": "ME",
    "mean_poisson_deviance": "MPD",
    "mean_gamma_deviance": "MGD",
    "mean_tweedie_deviance": "MTD",
    "continuous_ranked_probability_score": "CRPS",
}

units = {"wind": "kts", "pressure": "hPa"}


# %% Utilities
def _check_regression(y_true, y_pred):
    """Check that the regression inputs are of the correct form."""
    raise NotImplementedError
    assert y_true.shape == y_pred.shape, "y_true and y_pred must have the same shape."
    # assert y_true.ndim == 1, "y_true and y_pred must be 1D arrays."
    # assert y_pred.ndim == 1, "y_true and y_pred must be 1D arrays."
    return y_true, y_pred


def _check_classification(y_true, y_pred):
    """Check that the classification inputs are of the correct form."""
    raise NotImplementedError
    assert y_true.shape == y_pred.shape, "y_true and y_pred must have the same shape."
    # assert y_true.ndim == 1, "y_true and y_pred must be 1D arrays."
    # assert y_pred.ndim == 1, "y_true and y_pred must be 1D arrays."
    return y_true, y_pred


# %% Metrics


def CRPS(y_true, y_pred):
    """Compute the Continuous Ranked Probability Score (CRPS).

    The CRPS is a probabilistic metric that evaluates the accuracy of a
    probabilistic forecast. It is defined as the integral of the squared
    difference between the cumulative distribution function (CDF) of the
    forecast and the CDF of the observations.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        The true target values.

    y_pred : array-like of shape (n_samples, n_classes)
        The predicted probabilities for each class.

    Returns
    -------
    crps : float
        The CRPS value.
    """
    # Check that the inputs are of the correct form
    y_true, y_pred = _check_classification(y_true, y_pred)

    # Compute the CRPS
    raise NotImplementedError
    return crps


def CRPS_ML(y_pred, y_true, **kwargs):
    """Compute the Continuous Ranked Probability Score (CRPS).

    The CRPS is a probabilistic metric that evaluates the accuracy of a
    probabilistic forecast. It is defined as the integral of the squared
    difference between the cumulative distribution function (CDF) of the
    forecast and the CDF of the observations.

    Parameters
    ----------

    y_pred : array-like of shape (n_samples, mu, sigma)
        The predicted probability parameters for each sample.

    y_true : array-like of shape (n_samples,)

    Returns
    -------
    crps : float or array-like
        The CRPS value mean, or array of individual CRPS values.

    """
    # taken from https://github.com/WillyChap/ARML_Probabilistic/blob/main/Coastal_Points/Testing_and_Utility_Notebooks/CRPS_Verify.ipynb
    # and work by Louis Poulain--Auzeau (https://github.com/louisPoulain)
    reduction = kwargs.get("reduction", "mean")

    mu = y_pred[:, 0]
    sigma = y_pred[:, 1]

    # prevent negative sigmas
    sigma = torch.sqrt(sigma.pow(2))

    loc = (y_true - mu) / sigma
    pdf = torch.exp(-0.5 * loc.pow(2)) / torch.sqrt(
        2 * torch.from_numpy(np.array(np.pi))
    )
    cdf = 0.5 * (1.0 + torch.erf(loc / torch.sqrt(torch.tensor(2.0))))

    # compute CRPS for each (input, truth) pair
    crps = sigma * (
        loc * (2.0 * cdf - 1.0)
        + 2.0 * pdf
        - 1.0 / torch.from_numpy(np.array(np.sqrt(np.pi)))
    )

    return crps.mean() if reduction == "mean" else crps


def CRPSLoss(y_pred, y_true, **kwargs):
    reduction = kwargs.get("reduction", "mean")
    loss = 0.0
    indiv_losses = np.empty((y_true.shape))

    for i in range(0, y_true.shape[1]):
        l = CRPS_ML(y_pred[:, i * 2 : (i + 1) * 2], y_true[:, i], reduction=reduction)
        loss += l
        indiv_losses[:, i] = l.detach().cpu().numpy()

    return loss / (y_true.shape[1] // 2), indiv_losses


def CRPS_np(mu, sigma, y, **kwargs):
    reduction = kwargs.get("reduction", "mean")
    sigma = np.sqrt(sigma**2)
    loc = (y - mu) / sigma
    pdf = np.exp(-0.5 * loc**2) / np.sqrt(2 * np.pi)
    cdf = 0.5 * (1.0 + erf(loc / np.sqrt(2)))
    crps = sigma * (loc * (2.0 * cdf - 1.0) + 2.0 * pdf - 1.0 / np.sqrt(np.pi))
    return crps.mean() if reduction == "mean" else crps


def CRPSNumpy(mu_pred, sigma_pred, y_true, **kwargs):
    reduction = kwargs.get("reduction", "mean")
    indiv_losses = np.empty((y_true.shape))

    for i in range(0, y_true.shape[1]):
        l = CRPS_np(mu_pred[:, i], sigma_pred[:, i], y_true[:, i], reduction=reduction)
        indiv_losses[:, i] = l

    return indiv_losses


def summarize_performance(y_true, y_pred, y_baseline, metrics: list, **kwargs):
    """Summarize the performance of the model and the baseline.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        The true target values.

    y_pred : array-like of shape (n_samples,)
        The predicted target values.

    y_baseline : array-like of shape (n_samples,)
        The baseline target values.

    metrics : list of functions
        The list of metrics to compute.

    Returns
    -------
    performance : dict
        A dictionary containing the performance metrics.
    """
    # # Check that the inputs are of the correct form
    # y_true, y_pred = _check_regression(y_true, y_pred)
    # y_true, y_baseline = _check_regression(y_true, y_baseline)

    # Assert that the predictions have the shape (n_samples, n_features)
    assert y_pred.ndim == 2, "y_pred must have shape (n_samples, n_features)"

    y_labels = kwargs.get("y_labels", {0: "Wind", 1: "Pressure"})

    # Compute the performance metrics
    performance = {}
    for metric in metrics:
        for i in range(y_pred.shape[1]):
            y_true_i = y_true[:, i]
            y_pred_i = y_pred[:, i]
            y_baseline_i = y_baseline[:, i]

            if metric.__name__ in short_names.keys():
                metric_name = short_names[metric.__name__]
            else:
                metric_name = metric.__name__

            performance[f"{metric_name}_{y_labels[i]}"] = metric(y_true_i, y_pred_i)
            performance[f"{metric_name}_{y_labels[i]}_baseline"] = metric(
                y_true_i, y_baseline_i
            )

            if kwargs.get("skill", True):
                performance[f"{metric_name}_{y_labels[i]}_skill"] = (
                    1
                    - performance[f"{metric_name}_{y_labels[i]}"]
                    / performance[f"{metric_name}_{y_labels[i]}_baseline"]
                )
        # performance[metric.__name__] = metric(y_true, y_pred)
        # performance[f"{metric.__name__}_baseline"] = metric(y_true, y_baseline)

        # if kwargs.get("skill", True):
        #     performance[f"{metric.__name__}_skill"] = 1 - (
        #         performance[metric.__name__]
        #         / performance[f"{metric.__name__}_baseline"]
        #     )

    return performance


# %%
def plot_performance(metrics: dict, ax, **kwargs):
    """Plot the performance metrics.

    Parameters
    ----------
    metrics : dict
        A dictionary containing the performance metrics.

    ax : matplotlib.axes.Axes
        The axes to plot the metrics on.

    Returns
    -------
    None
    """
    if kwargs.get("skill", True):
        assert isinstance(
            ax, np.ndarray
        ), "ax should be a numpy array if skill=True (default)"

        ax1 = ax[0]
        ax2 = ax[1]
    else:
        ax1 = ax

    model_name = kwargs.get("model_name", "Model")
    baseline_name = kwargs.get("baseline_name", "Baseline")

    # Generate a list of unique metrics
    metric_names = []
    for metric in metrics.keys():
        if not ("skill" in metric or "baseline" in metric):
            metric_names.append(metric)
    metric_names = np.unique(metric_names)

    # Define the colors for the bars
    colors = kwargs.get("colors", plt.cm.tab20.colors)

    ax1_labels = []
    # Plot the performance metrics
    for i, metric in enumerate(metric_names):
        model_metric = metrics[metric]
        baseline_metric = metrics[f"{metric}_baseline"]
        var = metric.split("_")[-1].lower()
        unit = units.get(var, "")
        ax1.bar(
            [i * 3],
            [model_metric],
            color=colors[i % len(colors)],
            label=f"{model_name} {metric} ({unit})",
            hatch=kwargs.get("model_hatch", None),
        )

        ax1.bar(
            [i * 3 + 1],
            [baseline_metric],
            color=colors[i % len(colors)],
            label=f"{baseline_name} {metric} ({unit})",
            hatch=kwargs.get("baseline_hatch", "//"),
        )

        ax1.set_ylabel("Score")
        ax1.set_title("Performance Metrics")

        ax1_labels += [f"{model_name} metric", f"{baseline_name} {metric}", ""]

        if kwargs.get("skill", True):
            ax2.bar(
                [i],
                [metrics[f"{metric}_skill"]],
                color=colors[i % len(colors)],
                label=f"{metric} Skill Score",
                hatch=kwargs.get("skill_hatch", None),
            )
            ax2.set_ylabel("Skill Score (1 - model/baseline)")
            ax2.set_title("Skill Scores")

    ax1.set_xticks(range(len(metric_names) * 3))
    ax1.set_xticklabels([""] * len(ax1_labels))
    ax1.legend(loc="lower right", framealpha=0.5)

    if kwargs.get("skill", True):
        ax2.set_xticks(range(len(metric_names)))
        ax2.set_xticklabels([""] * len(metric_names))
        ax2.legend(loc="lower right", framealpha=0.5)
